# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from .bigtable_instance_admin import (
    CreateAppProfileRequest,
    CreateClusterMetadata,
    CreateClusterRequest,
    CreateInstanceMetadata,
    CreateInstanceRequest,
    CreateLogicalViewMetadata,
    CreateLogicalViewRequest,
    CreateMaterializedViewMetadata,
    CreateMaterializedViewRequest,
    DeleteAppProfileRequest,
    DeleteClusterRequest,
    DeleteInstanceRequest,
    DeleteLogicalViewRequest,
    DeleteMaterializedViewRequest,
    GetAppProfileRequest,
    GetClusterRequest,
    GetInstanceRequest,
    GetLogicalViewRequest,
    GetMaterializedViewRequest,
    ListAppProfilesRequest,
    ListAppProfilesResponse,
    ListClustersRequest,
    ListClustersResponse,
    ListHotTabletsRequest,
    ListHotTabletsResponse,
    ListInstancesRequest,
    ListInstancesResponse,
    ListLogicalViewsRequest,
    ListLogicalViewsResponse,
    ListMaterializedViewsRequest,
    ListMaterializedViewsResponse,
    PartialUpdateClusterMetadata,
    PartialUpdateClusterRequest,
    PartialUpdateInstanceRequest,
    UpdateAppProfileMetadata,
    UpdateAppProfileRequest,
    UpdateClusterMetadata,
    UpdateInstanceMetadata,
    UpdateLogicalViewMetadata,
    UpdateLogicalViewRequest,
    UpdateMaterializedViewMetadata,
    UpdateMaterializedViewRequest,
)
from .bigtable_table_admin import (
    CheckConsistencyRequest,
    CheckConsistencyResponse,
    CopyBackupMetadata,
    CopyBackupRequest,
    CreateAuthorizedViewMetadata,
    CreateAuthorizedViewRequest,
    CreateBackupMetadata,
    CreateBackupRequest,
    CreateSchemaBundleMetadata,
    CreateSchemaBundleRequest,
    CreateTableFromSnapshotMetadata,
    CreateTableFromSnapshotRequest,
    CreateTableRequest,
    DataBoostReadLocalWrites,
    DeleteAuthorizedViewRequest,
    DeleteBackupRequest,
    DeleteSchemaBundleRequest,
    DeleteSnapshotRequest,
    DeleteTableRequest,
    DropRowRangeRequest,
    GenerateConsistencyTokenRequest,
    GenerateConsistencyTokenResponse,
    GetAuthorizedViewRequest,
    GetBackupRequest,
    GetSchemaBundleRequest,
    GetSnapshotRequest,
    GetTableRequest,
    ListAuthorizedViewsRequest,
    ListAuthorizedViewsResponse,
    ListBackupsRequest,
    ListBackupsResponse,
    ListSchemaBundlesRequest,
    ListSchemaBundlesResponse,
    ListSnapshotsRequest,
    ListSnapshotsResponse,
    ListTablesRequest,
    ListTablesResponse,
    ModifyColumnFamiliesRequest,
    OptimizeRestoredTableMetadata,
    RestoreTableMetadata,
    RestoreTableRequest,
    SnapshotTableMetadata,
    SnapshotTableRequest,
    StandardReadRemoteWrites,
    UndeleteTableMetadata,
    UndeleteTableRequest,
    UpdateAuthorizedViewMetadata,
    UpdateAuthorizedViewRequest,
    UpdateBackupRequest,
    UpdateSchemaBundleMetadata,
    UpdateSchemaBundleRequest,
    UpdateTableMetadata,
    UpdateTableRequest,
)
from .common import (
    OperationProgress,
    StorageType,
)
from .instance import (
    AppProfile,
    AutoscalingLimits,
    AutoscalingTargets,
    Cluster,
    HotTablet,
    Instance,
    LogicalView,
    MaterializedView,
)
from .table import (
    AuthorizedView,
    Backup,
    BackupInfo,
    ChangeStreamConfig,
    ColumnFamily,
    EncryptionInfo,
    GcRule,
    ProtoSchema,
    RestoreInfo,
    SchemaBundle,
    Snapshot,
    Table,
    RestoreSourceType,
)
from .types import (
    Type,
)

__all__ = (
    "CreateAppProfileRequest",
    "CreateClusterMetadata",
    "CreateClusterRequest",
    "CreateInstanceMetadata",
    "CreateInstanceRequest",
    "CreateLogicalViewMetadata",
    "CreateLogicalViewRequest",
    "CreateMaterializedViewMetadata",
    "CreateMaterializedViewRequest",
    "DeleteAppProfileRequest",
    "DeleteClusterRequest",
    "DeleteInstanceRequest",
    "DeleteLogicalViewRequest",
    "DeleteMaterializedViewRequest",
    "GetAppProfileRequest",
    "GetClusterRequest",
    "GetInstanceRequest",
    "GetLogicalViewRequest",
    "GetMaterializedViewRequest",
    "ListAppProfilesRequest",
    "ListAppProfilesResponse",
    "ListClustersRequest",
    "ListClustersResponse",
    "ListHotTabletsRequest",
    "ListHotTabletsResponse",
    "ListInstancesRequest",
    "ListInstancesResponse",
    "ListLogicalViewsRequest",
    "ListLogicalViewsResponse",
    "ListMaterializedViewsRequest",
    "ListMaterializedViewsResponse",
    "PartialUpdateClusterMetadata",
    "PartialUpdateClusterRequest",
    "PartialUpdateInstanceRequest",
    "UpdateAppProfileMetadata",
    "UpdateAppProfileRequest",
    "UpdateClusterMetadata",
    "UpdateInstanceMetadata",
    "UpdateLogicalViewMetadata",
    "UpdateLogicalViewRequest",
    "UpdateMaterializedViewMetadata",
    "UpdateMaterializedViewRequest",
    "CheckConsistencyRequest",
    "CheckConsistencyResponse",
    "CopyBackupMetadata",
    "CopyBackupRequest",
    "CreateAuthorizedViewMetadata",
    "CreateAuthorizedViewRequest",
    "CreateBackupMetadata",
    "CreateBackupRequest",
    "CreateSchemaBundleMetadata",
    "CreateSchemaBundleRequest",
    "CreateTableFromSnapshotMetadata",
    "CreateTableFromSnapshotRequest",
    "CreateTableRequest",
    "DataBoostReadLocalWrites",
    "DeleteAuthorizedViewRequest",
    "DeleteBackupRequest",
    "DeleteSchemaBundleRequest",
    "DeleteSnapshotRequest",
    "DeleteTableRequest",
    "DropRowRangeRequest",
    "GenerateConsistencyTokenRequest",
    "GenerateConsistencyTokenResponse",
    "GetAuthorizedViewRequest",
    "GetBackupRequest",
    "GetSchemaBundleRequest",
    "GetSnapshotRequest",
    "GetTableRequest",
    "ListAuthorizedViewsRequest",
    "ListAuthorizedViewsResponse",
    "ListBackupsRequest",
    "ListBackupsResponse",
    "ListSchemaBundlesRequest",
    "ListSchemaBundlesResponse",
    "ListSnapshotsRequest",
    "ListSnapshotsResponse",
    "ListTablesRequest",
    "ListTablesResponse",
    "ModifyColumnFamiliesRequest",
    "OptimizeRestoredTableMetadata",
    "RestoreTableMetadata",
    "RestoreTableRequest",
    "SnapshotTableMetadata",
    "SnapshotTableRequest",
    "StandardReadRemoteWrites",
    "UndeleteTableMetadata",
    "UndeleteTableRequest",
    "UpdateAuthorizedViewMetadata",
    "UpdateAuthorizedViewRequest",
    "UpdateBackupRequest",
    "UpdateSchemaBundleMetadata",
    "UpdateSchemaBundleRequest",
    "UpdateTableMetadata",
    "UpdateTableRequest",
    "OperationProgress",
    "StorageType",
    "AppProfile",
    "AutoscalingLimits",
    "AutoscalingTargets",
    "Cluster",
    "HotTablet",
    "Instance",
    "LogicalView",
    "MaterializedView",
    "AuthorizedView",
    "Backup",
    "BackupInfo",
    "ChangeStreamConfig",
    "ColumnFamily",
    "EncryptionInfo",
    "GcRule",
    "ProtoSchema",
    "RestoreInfo",
    "SchemaBundle",
    "Snapshot",
    "Table",
    "RestoreSourceType",
    "Type",
)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""User friendly container for Google Cloud Bigtable MutationBatcher."""
import threading
import queue
import concurrent.futures
import atexit


from google.api_core.exceptions import from_grpc_status
from dataclasses import dataclass


FLUSH_COUNT = 100  # after this many elements, send out the batch

MAX_MUTATION_SIZE = 20 * 1024 * 1024  # 20MB # after this many bytes, send out the batch

MAX_OUTSTANDING_BYTES = 100 * 1024 * 1024  # 100MB # max inflight byte size.

MAX_OUTSTANDING_ELEMENTS = 100000  # max inflight mutations.


class MutationsBatchError(Exception):
    """Error in the batch request"""

    def __init__(self, message, exc):
        self.exc = exc
        self.message = message
        super().__init__(self.message)


class _MutationsBatchQueue(object):
    """Private Threadsafe Queue to hold rows for batching."""

    def __init__(self, max_mutation_bytes=MAX_MUTATION_SIZE, flush_count=FLUSH_COUNT):
        """Specify the queue constraints"""
        self._queue = queue.Queue()
        self.total_mutation_count = 0
        self.total_size = 0
        self.max_mutation_bytes = max_mutation_bytes
        self.flush_count = flush_count

    def get(self):
        """
        Retrieve an item from the queue. Recalculate queue size.

        If the queue is empty, return None.
        """
        try:
            row = self._queue.get_nowait()
            mutation_size = row.get_mutations_size()
            self.total_mutation_count -= len(row._get_mutations())
            self.total_size -= mutation_size
            return row
        except queue.Empty:
            return None

    def put(self, item):
        """Insert an item to the queue. Recalculate queue size."""

        mutation_count = len(item._get_mutations())

        self._queue.put(item)

        self.total_size += item.get_mutations_size()
        self.total_mutation_count += mutation_count

    def full(self):
        """Check if the queue is full."""
        if (
            self.total_mutation_count >= self.flush_count
            or self.total_size >= self.max_mutation_bytes
        ):
            return True
        return False


@dataclass
class _BatchInfo:
    """Keeping track of size of a batch"""

    mutations_count: int = 0
    rows_count: int = 0
    mutations_size: int = 0


class _FlowControl(object):
    def __init__(
        self,
        max_mutations=MAX_OUTSTANDING_ELEMENTS,
        max_mutation_bytes=MAX_OUTSTANDING_BYTES,
    ):
        """Control the inflight requests. Keep track of the mutations, row bytes and row counts.
        As requests to backend are being made, adjust the number of mutations being processed.

        If threshold is reached, block the flow.
        Reopen the flow as requests are finished.
        """
        self.max_mutations = max_mutations
        self.max_mutation_bytes = max_mutation_bytes
        self.inflight_mutations = 0
        self.inflight_size = 0
        self.event = threading.Event()
        self.event.set()
        self._lock = threading.Lock()

    def is_blocked(self):
        """Returns True if:

        - inflight mutations >= max_mutations, or
        - inflight bytes size >= max_mutation_bytes, or
        """

        return (
            self.inflight_mutations >= self.max_mutations
            or self.inflight_size >= self.max_mutation_bytes
        )

    def control_flow(self, batch_info):
        """
        Calculate the resources used by this batch
        """

        with self._lock:
            self.inflight_mutations += batch_info.mutations_count
            self.inflight_size += batch_info.mutations_size
        self.set_flow_control_status()

    def wait(self):
        """
        Wait until flow control pushback has been released.
        It awakens as soon as `event` is set.
        """
        self.event.wait()

    def set_flow_control_status(self):
        """Check the inflight mutations and size.

        If values exceed the allowed threshold, block the event.
        """
        if self.is_blocked():
            self.event.clear()  # sleep
        else:
            self.event.set()  # awaken the threads

    def release(self, batch_info):
        """
        Release the resources.
        Decrement the row size to allow enqueued mutations to be run.
        """
        with self._lock:
            self.inflight_mutations -= batch_info.mutations_count
            self.inflight_size -= batch_info.mutations_size
        self.set_flow_control_status()


class MutationsBatcher(object):
    """A MutationsBatcher is used in batch cases where the number of mutations
    is large or unknown. It will store :class:`DirectRow` in memory until one of the
    size limits is reached, or an explicit call to :func:`flush()` is performed. When
    a flush event occurs, the :class:`DirectRow` in memory will be sent to Cloud
    Bigtable. Batching mutations is more efficient than sending individual
    request.

    This class is not suited for usage in systems where each mutation
    must be guaranteed to be sent, since calling mutate may only result in an
    in-memory change. In a case of a system crash, any :class:`DirectRow` remaining in
    memory will not necessarily be sent to the service, even after the
    completion of the :func:`mutate()` method.

    Note on thread safety: The same :class:`MutationBatcher` cannot be shared by multiple end-user threads.

    :type table: class
    :param table: class:`~google.cloud.bigtable.table.Table`.

    :type flush_count: int
    :param flush_count: (Optional) Max number of rows to flush. If it
        reaches the max number of rows it calls finish_batch() to mutate the
        current row batch. Default is FLUSH_COUNT (1000 rows).

    :type max_row_bytes: int
    :param max_row_bytes: (Optional) Max number of row mutations size to
        flush. If it reaches the max number of row mutations size it calls
        finish_batch() to mutate the current row batch. Default is MAX_ROW_BYTES
        (5 MB).

    :type flush_interval: float
    :param flush_interval: (Optional) The interval (in seconds) between asynchronous flush.
        Default is 1 second.

    :type batch_completed_callback: Callable[list:[`~google.rpc.status_pb2.Status`]] = None
    :param batch_completed_callback: (Optional) A callable for handling responses
        after the current batch is sent. The callable function expect a list of grpc
        Status.
    """

    def __init__(
        self,
        table,
        flush_count=FLUSH_COUNT,
        max_row_bytes=MAX_MUTATION_SIZE,
        flush_interval=1,
        batch_completed_callback=None,
    ):
        self._rows = _MutationsBatchQueue(
            max_mutation_bytes=max_row_bytes, flush_count=flush_count
        )
        self.table = table
        self._executor = concurrent.futures.ThreadPoolExecutor()
        atexit.register(self.close)
        self._timer = threading.Timer(flush_interval, self.flush)
        self._timer.start()
        self.flow_control = _FlowControl(
            max_mutations=MAX_OUTSTANDING_ELEMENTS,
            max_mutation_bytes=MAX_OUTSTANDING_BYTES,
        )
        self.futures_mapping = {}
        self.exceptions = queue.Queue()
        self._user_batch_completed_callback = batch_completed_callback

    @property
    def flush_count(self):
        return self._rows.flush_count

    @property
    def max_row_bytes(self):
        return self._rows.max_mutation_bytes

    def __enter__(self):
        """Starting the MutationsBatcher as a context manager"""
        return self

    def mutate(self, row):
        """Add a row to the batch. If the current batch meets one of the size
        limits, the batch is sent asynchronously.

        For example:

        .. literalinclude:: snippets_table.py
            :start-after: [START bigtable_api_batcher_mutate]
            :end-before: [END bigtable_api_batcher_mutate]
            :dedent: 4

        :type row: class
        :param row: :class:`~google.cloud.bigtable.row.DirectRow`.

        :raises: One of the following:
            * :exc:`~.table._BigtableRetryableError` if any row returned a transient error.
            * :exc:`RuntimeError` if the number of responses doesn't match the number of rows that were retried
        """
        self._rows.put(row)

        if self._rows.full():
            self._flush_async()

    def mutate_rows(self, rows):
        """Add multiple rows to the batch. If the current batch meets one of the size
        limits, the batch is sent asynchronously.

        For example:

        .. literalinclude:: snippets_table.py
            :start-after: [START bigtable_api_batcher_mutate_rows]
            :end-before: [END bigtable_api_batcher_mutate_rows]
            :dedent: 4

        :type rows: list:[`~google.cloud.bigtable.row.DirectRow`]
        :param rows: list:[`~google.cloud.bigtable.row.DirectRow`].

        :raises: One of the following:
            * :exc:`~.table._BigtableRetryableError` if any row returned a transient error.
            * :exc:`RuntimeError` if the number of responses doesn't match the number of rows that were retried
        """
        for row in rows:
            self.mutate(row)

    def flush(self):
        """Sends the current batch to Cloud Bigtable synchronously.
        For example:

        .. literalinclude:: snippets_table.py
            :start-after: [START bigtable_api_batcher_flush]
            :end-before: [END bigtable_api_batcher_flush]
            :dedent: 4

        :raises:
            * :exc:`.batcherMutationsBatchError` if there's any error in the mutations.
        """
        rows_to_flush = []
        row = self._rows.get()
        while row is not None:
            rows_to_flush.append(row)
            row = self._rows.get()
        response = self._flush_rows(rows_to_flush)
        return response

    def _flush_async(self):
        """Sends the current batch to Cloud Bigtable asynchronously.

        :raises:
            * :exc:`.batcherMutationsBatchError` if there's any error in the mutations.
        """
        next_row = self._rows.get()
        while next_row is not None:
            # start a new batch
            rows_to_flush = [next_row]
            batch_info = _BatchInfo(
                mutations_count=len(next_row._get_mutations()),
                rows_count=1,
                mutations_size=next_row.get_mutations_size(),
            )
            # fill up batch with rows
            next_row = self._rows.get()
            while next_row is not None and self._row_fits_in_batch(
                next_row, batch_info
            ):
                rows_to_flush.append(next_row)
                batch_info.mutations_count += len(next_row._get_mutations())
                batch_info.rows_count += 1
                batch_info.mutations_size += next_row.get_mutations_size()
                next_row = self._rows.get()
            # send batch over network
            # wait for resources to become available
            self.flow_control.wait()
            # once unblocked, submit the batch
            # event flag will be set by control_flow to block subsequent thread, but not blocking this one
            self.flow_control.control_flow(batch_info)
            future = self._executor.submit(self._flush_rows, rows_to_flush)
            # schedule release of resources from flow control
            self.futures_mapping[future] = batch_info
            future.add_done_callback(self._batch_completed_callback)

    def _batch_completed_callback(self, future):
        """Callback for when the mutation has finished to clean up the current batch
        and release items from the flow controller.
        Raise exceptions if there's any.
        Release the resources locked by the flow control and allow enqueued tasks to be run.
        """
        processed_rows = self.futures_mapping[future]
        self.flow_control.release(processed_rows)
        del self.futures_mapping[future]

    def _row_fits_in_batch(self, row, batch_info):
        """Checks if a row can fit in the current batch.

        :type row: class
        :param row: :class:`~google.cloud.bigtable.row.DirectRow`.

        :type batch_info: :class:`_BatchInfo`
        :param batch_info: Information about the current batch.

        :rtype: bool
        :returns: True if the row can fit in the current batch.
        """
        new_rows_count = batch_info.rows_count + 1
        new_mutations_count = batch_info.mutations_count + len(row._get_mutations())
        new_mutations_size = batch_info.mutations_size + row.get_mutations_size()
        return (
            new_rows_count <= self.flush_count
            and new_mutations_size <= self.max_row_bytes
            and new_mutations_count <= self.flow_control.max_mutations
            and new_mutations_size <= self.flow_control.max_mutation_bytes
        )

    def _flush_rows(self, rows_to_flush):
        """Mutate the specified rows.

        :raises:
            * :exc:`.batcherMutationsBatchError` if there's any error in the mutations.
        """
        responses = []
        if len(rows_to_flush) > 0:
            response = self.table.mutate_rows(rows_to_flush)

            if self._user_batch_completed_callback:
                self._user_batch_completed_callback(response)

            for result in response:
                if result.code != 0:
                    exc = from_grpc_status(result.code, result.message)
                    self.exceptions.put(exc)
                responses.append(result)

        return responses

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Clean up resources. Flush and shutdown the ThreadPoolExecutor."""
        self.close()

    def close(self):
        """Clean up resources. Flush and shutdown the ThreadPoolExecutor.
        Any errors will be raised.

        :raises:
            * :exc:`.batcherMutationsBatchError` if there's any error in the mutations.
        """
        self.flush()
        self._executor.shutdown(wait=True)
        atexit.unregister(self.close)
        if self.exceptions.qsize() > 0:
            exc = list(self.exceptions.queue)
            raise MutationsBatchError("Errors in batch mutations.", exc=exc)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Class to support Batch Prediction with GenAI models."""
# pylint: disable=protected-access

import logging
import re
from typing import List, Optional, Union

from google.cloud.aiplatform import base as aiplatform_base
from google.cloud.aiplatform import initializer as aiplatform_initializer
from google.cloud.aiplatform import jobs
from google.cloud.aiplatform import models
from google.cloud.aiplatform import utils as aiplatform_utils
from google.cloud.aiplatform_v1 import types as gca_types
from vertexai import generative_models

from google.rpc import status_pb2


_LOGGER = aiplatform_base.Logger(__name__)

_GEMINI_MODEL_PATTERN = r"publishers/google/models/gemini"
_LLAMA_MODEL_PATTERN = r"publishers/meta/models/llama"
_CLAUDE_MODEL_PATTERN = r"publishers/anthropic/models/claude"
_GPT_MODEL_PATTERN = r"publishers/openai/models/gpt"
_QWEN_MODEL_PATTERN = r"publishers/qwen/models/qwen"
_DEEPSEEK_MODEL_PATTERN = r"publishers/deepseek-ai/models/deepseek"
_E5_MODEL_PATTERN = r"publishers/intfloat/models/multilingual"
_GEMINI_TUNED_MODEL_PATTERN = r"^projects/[0-9]+?/locations/[0-9a-z-]+?/models/[0-9]+?$"


class BatchPredictionJob(aiplatform_base._VertexAiResourceNounPlus):
    """Represents a BatchPredictionJob that runs with GenAI models."""

    _resource_noun = "batchPredictionJobs"
    _getter_method = "get_batch_prediction_job"
    _list_method = "list_batch_prediction_jobs"
    _delete_method = "delete_batch_prediction_job"
    _job_type = "batch-predictions"
    _parse_resource_name_method = "parse_batch_prediction_job_path"
    _format_resource_name_method = "batch_prediction_job_path"

    client_class = aiplatform_utils.JobClientWithOverride

    def __init__(self, batch_prediction_job_name: str):
        """Retrieves a BatchPredictionJob resource that runs with a GenAI model.

        Args:
            batch_prediction_job_name (str):
                Required. A fully-qualified BatchPredictionJob resource name or
                ID. Example: "projects/.../locations/.../batchPredictionJobs/456"
                or "456" when project and location are initialized.

        Raises:
            ValueError: If batch_prediction_job_name represents a BatchPredictionJob
            resource that runs with another type of model.
        """
        super().__init__(resource_name=batch_prediction_job_name)
        self._gca_resource = self._get_gca_resource(
            resource_name=batch_prediction_job_name
        )
        if not self._is_genai_model(self.model_name):
            raise ValueError(
                f"BatchPredictionJob '{batch_prediction_job_name}' "
                f"runs with the model '{self.model_name}', "
                "which is not a GenAI model."
            )

    @property
    def model_name(self) -> str:
        """Returns the model name used for this batch prediction job."""
        return self._gca_resource.model

    @property
    def state(self) -> gca_types.JobState:
        """Returns the state of this batch prediction job."""
        return self._gca_resource.state

    @property
    def has_ended(self) -> bool:
        """Returns true if this batch prediction job has ended."""
        return self.state in jobs._JOB_COMPLETE_STATES

    @property
    def has_succeeded(self) -> bool:
        """Returns true if this batch prediction job has succeeded."""
        return self.state == gca_types.JobState.JOB_STATE_SUCCEEDED

    @property
    def error(self) -> Optional[status_pb2.Status]:
        """Returns detailed error info for this Job resource."""
        return self._gca_resource.error

    @property
    def output_location(self) -> str:
        """Returns the output location of this batch prediction job."""
        return (
            self._gca_resource.output_info.gcs_output_directory
            or self._gca_resource.output_info.bigquery_output_table
        )

    @classmethod
    def submit(
        cls,
        source_model: Union[str, generative_models.GenerativeModel],
        input_dataset: Union[str, List[str]],
        *,
        output_uri_prefix: Optional[str] = None,
        job_display_name: Optional[str] = None,
        machine_type: Optional[str] = None,
        accelerator_type: Optional[str] = None,
        accelerator_count: Optional[int] = None,
        starting_replica_count: Optional[int] = None,
        max_replica_count: Optional[int] = None,
    ) -> "BatchPredictionJob":
        """Submits a batch prediction job for a GenAI model.

        Args:
            source_model (Union[str, generative_models.GenerativeModel]):
                A GenAI model name or a tuned model name or a GenerativeModel instance
                for batch prediction.
                Supported formats for model name: "gemini-1.0-pro",
                "models/gemini-1.0-pro", and "publishers/google/models/gemini-1.0-pro"
                Supported formats for tuned model name: "789" and
                "projects/123/locations/456/models/789"
            input_dataset (Union[str,List[str]]):
                GCS URI(-s) or BigQuery URI to your input data to run batch
                prediction on. Example: "gs://path/to/input/data.jsonl" or
                "bq://projectId.bqDatasetId.bqTableId"
            output_uri_prefix (str):
                GCS or BigQuery URI prefix for the output predictions. Example:
                "gs://path/to/output/data" or "bq://projectId.bqDatasetId"
                If not specified, f"{STAGING_BUCKET}/gen-ai-batch-prediction" will
                be used for GCS source and
                f"bq://projectId.gen_ai_batch_prediction.predictions_{TIMESTAMP}"
                will be used for BigQuery source.
            job_display_name (str):
                The user-defined name of the BatchPredictionJob.
                The name can be up to 128 characters long and can be consist
                of any UTF-8 characters.
            machine_type (str):
                The type of machine for running batch prediction job.
            accelerator_type (str):
                The type of accelerator for running batch prediction job.
            accelerator_count (int):
                The number of accelerators for running batch prediction job.
            starting_replica_count (int):
                The starting number of replica for running batch prediction job.
            max_replica_count (int):
                The maximum number of replica for running batch prediction job.

        Returns:
            Instantiated BatchPredictionJob.

        Raises:
            ValueError: If source_model is not a GenAI model.
            Or if input_dataset or output_uri_prefix are not in supported formats.
            Or if output_uri_prefix is not specified and staging_bucket is not
            set in vertexai.init().
        """
        # Handle model name
        model_name = cls._reconcile_model_name(
            source_model._model_name
            if isinstance(source_model, generative_models.GenerativeModel)
            else source_model
        )
        if not cls._is_genai_model(model_name):
            raise ValueError(f"Model '{model_name}' is not a Generative AI model.")

        # Handle input URI
        gcs_source = None
        bigquery_source = None
        first_input_uri = (
            input_dataset if isinstance(input_dataset, str) else input_dataset[0]
        )
        if first_input_uri.startswith("gs://"):
            gcs_source = input_dataset
        elif first_input_uri.startswith("bq://"):
            if not isinstance(input_dataset, str):
                raise ValueError("Multiple BigQuery input datasets are not supported.")
            bigquery_source = input_dataset
        else:
            raise ValueError(
                f"Unsupported input URI: {input_dataset}. "
                "Supported formats: 'gs://path/to/input/data.jsonl' and "
                "'bq://projectId.bqDatasetId.bqTableId'"
            )

        # Handle output URI
        gcs_destination_prefix = None
        bigquery_destination_prefix = None
        if output_uri_prefix:
            if output_uri_prefix.startswith("gs://"):
                gcs_destination_prefix = output_uri_prefix
            elif output_uri_prefix.startswith("bq://"):
                # Temporarily handle this in SDK, will remove once b/338423462 is fixed.
                bigquery_destination_prefix = cls._complete_bq_uri(output_uri_prefix)
            else:
                raise ValueError(
                    f"Unsupported output URI: {output_uri_prefix}. "
                    "Supported formats: 'gs://path/to/output/data' and "
                    "'bq://projectId.bqDatasetId'"
                )
        else:
            if first_input_uri.startswith("gs://"):
                if not aiplatform_initializer.global_config.staging_bucket:
                    raise ValueError(
                        "Please either specify output_uri_prefix or "
                        "set staging_bucket in vertexai.init()."
                    )
                gcs_destination_prefix = (
                    aiplatform_initializer.global_config.staging_bucket.rstrip("/")
                    + "/gen-ai-batch-prediction"
                )
            else:
                bigquery_destination_prefix = cls._complete_bq_uri()

        # Reuse aiplatform class to submit the job (override _LOGGER)
        logging.getLogger("google.cloud.aiplatform.jobs").disabled = True
        try:
            aiplatform_job = jobs.BatchPredictionJob.submit(
                model_name=model_name,
                job_display_name=job_display_name,
                gcs_source=gcs_source,
                bigquery_source=bigquery_source,
                gcs_destination_prefix=gcs_destination_prefix,
                bigquery_destination_prefix=bigquery_destination_prefix,
                machine_type=machine_type,
                accelerator_type=accelerator_type,
                accelerator_count=accelerator_count,
                starting_replica_count=starting_replica_count,
                max_replica_count=max_replica_count,
            )
            job = cls._empty_constructor()
            job._gca_resource = aiplatform_job._gca_resource

            _LOGGER.log_create_complete(
                cls, job._gca_resource, "job", module_name="batch_prediction"
            )
            _LOGGER.info("View Batch Prediction Job:\n%s" % job._dashboard_uri())

            return job
        finally:
            logging.getLogger("google.cloud.aiplatform.jobs").disabled = False

    def refresh(self) -> "BatchPredictionJob":
        """Refreshes the batch prediction job from the service."""
        self._sync_gca_resource()
        return self

    def cancel(self):
        """Cancels this BatchPredictionJob.

        Success of cancellation is not guaranteed. Use `job.refresh()` and
        `job.state` to verify if cancellation was successful.
        """
        _LOGGER.log_action_start_against_resource("Cancelling", "run", self)
        self.api_client.cancel_batch_prediction_job(name=self.resource_name)

    def delete(self):
        """Deletes this BatchPredictionJob resource.

        WARNING: This deletion is permanent.
        """
        self._delete()

    @classmethod
    def list(cls, filter=None) -> List["BatchPredictionJob"]:
        """Lists all BatchPredictionJob instances that run with GenAI models."""
        return cls._list(
            cls_filter=lambda gca_resource: cls._is_genai_model(gca_resource.model),
            filter=filter,
        )

    def _dashboard_uri(self) -> Optional[str]:
        """Returns the Google Cloud console URL where job can be viewed."""
        fields = self._parse_resource_name(self.resource_name)
        location = fields.pop("location")
        project = fields.pop("project")
        job = list(fields.values())[0]
        return (
            "https://console.cloud.google.com/ai/platform/locations/"
            f"{location}/{self._job_type}/{job}?project={project}"
        )

    @classmethod
    def _reconcile_model_name(cls, model_name: str) -> str:
        """Reconciles model name to a publisher model resource name or a tuned model resource name."""
        if not model_name:
            raise ValueError("model_name must not be empty")

        if "/" not in model_name:
            # model name (e.g., gemini-1.0-pro)
            if model_name.startswith("gemini"):
                return "publishers/google/models/" + model_name
            else:
                raise ValueError(
                    "Abbreviated model names are only supported for Gemini models. "
                    "Please provide the full publisher model name."
                )
        elif model_name.startswith("models/"):
            # publisher model name (e.g., models/gemini-1.0-pro)
            return "publishers/google/" + model_name
        elif (
            re.match(
                r"^publishers/(?P<publisher>[^/]+)/models/(?P<model>[^@]+)@(?P<version>[^@]+)$",
                model_name,
            )
            or model_name.startswith("publishers/google/models/")
            or model_name.startswith("publishers/meta/models/")
            or model_name.startswith("publishers/anthropic/models/")
            or model_name.startswith("publishers/openai/models/")
            or model_name.startswith("publishers/qwen/models/")
            or model_name.startswith("publishers/deepseek-ai/models/")
            or model_name.startswith("publishers/intfloat/models/")
            or re.search(_GEMINI_TUNED_MODEL_PATTERN, model_name)
        ):
            return model_name
        else:
            raise ValueError(f"Invalid format for model name: {model_name}.")

    @classmethod
    def _is_genai_model(cls, model_name: str) -> bool:
        """Validates if a given model_name represents a GenAI model."""
        if re.search(_GEMINI_MODEL_PATTERN, model_name):
            # Model is a Gemini model.
            return True

        if re.search(_GEMINI_TUNED_MODEL_PATTERN, model_name):
            model = models.Model(model_name)
            if (
                model.gca_resource.model_source_info.source_type
                == gca_types.model.ModelSourceInfo.ModelSourceType.GENIE
            ):
                # Model is a tuned Gemini model.
                return True

        if re.search(_LLAMA_MODEL_PATTERN, model_name):
            # Model is a Llama3 model.
            return True

        if re.search(_CLAUDE_MODEL_PATTERN, model_name):
            # Model is a claude model.
            return True

        if re.search(_GPT_MODEL_PATTERN, model_name):
            # Model is a GPT model.
            return True

        if re.search(_QWEN_MODEL_PATTERN, model_name):
            # Model is a Qwen model.
            return True

        if re.search(_DEEPSEEK_MODEL_PATTERN, model_name):
            # Model is a DeepSeek model.
            return True

        if re.search(_E5_MODEL_PATTERN, model_name):
            # Model is an E5 model.
            return True

        if re.match(
            r"^publishers/(?P<publisher>[^/]+)/models/(?P<model>[^@]+)@(?P<version>[^@]+)$",
            model_name,
        ):
            # Model is a self-hosted model.
            return True

        return False

    @classmethod
    def num_pending_jobs(cls) -> int:
        """Returns the number of pending batch prediction jobs.

        The pending jobs are those defined in _JOB_PENDING_STATES from
        google/cloud/aiplatform/jobs.py
        e.g. JOB_STATE_QUEUED, JOB_STATE_PENDING, JOB_STATE_RUNNING,
        JOB_STATE_CANCELLING, JOB_STATE_UPDATING.
        It will be used to manage the number of concurrent batch that is limited
        according to
        https://cloud.google.com/vertex-ai/generative-ai/docs/quotas#concurrent-batch-requests
        """
        return len(
            cls._list(
                cls_filter=lambda gca_resource: cls._is_genai_model(gca_resource.model),
                filter=" OR ".join(
                    f'state="{pending_state.name}"'
                    for pending_state in jobs._JOB_PENDING_STATES
                ),
            )
        )

    @classmethod
    def _complete_bq_uri(cls, uri: Optional[str] = None):
        """Completes a BigQuery uri to a BigQuery table uri."""
        uri_parts = uri.split(".") if uri else []
        uri_len = len(uri_parts)
        if len(uri_parts) > 3:
            raise ValueError(
                f"Invalid URI: {uri}. "
                "Supported formats: 'bq://projectId.bqDatasetId.bqTableId'"
            )

        schema_and_project = (
            uri_parts[0]
            if uri_len >= 1
            else f"bq://{aiplatform_initializer.global_config.project}"
        )
        if not schema_and_project.startswith("bq://"):
            raise ValueError("URI must start with 'bq://'")

        dataset = uri_parts[1] if uri_len >= 2 else "gen_ai_batch_prediction"

        table = (
            uri_parts[2]
            if uri_len >= 3
            else f"predictions_{aiplatform_utils.timestamped_unique_name()}"
        )

        return f"{schema_and_project}.{dataset}.{table}"



"""
Enhanced Streamlit App for Gemini 2.5 Pro + BMasterAI + Tavily + Firecrawl

This Streamlit application demonstrates real-time BMasterAI reasoning transparency
by showing how Gemini thinks through complex tasks like finding AI podcast influencers
using Tavily API for search and Firecrawl API for email extraction.
"""

import streamlit as st
import os
import json
import glob
import time
from datetime import datetime
from typing import Dict, List, Any, Optional
import google.generativeai as genai

# Import BMasterAI components
try:
    from bmasterai import configure_logging, LogLevel
    BMASTERAI_AVAILABLE = True
except ImportError:
    BMASTERAI_AVAILABLE = False
    st.error("BMasterAI library not found. Please install it with: pip install bmasterai")

# Import our enhanced Gemini reasoning agent
from gemini_reasoning_app import GeminiReasoningAgent

# Load environment variables from .env file if it exists
try:
    from dotenv import load_dotenv
    load_dotenv()
except ImportError:
    st.warning("python-dotenv not installed. Install it with: pip install python-dotenv")

def load_css():
    """Load custom CSS for better styling"""
    st.markdown("""
    <style>
    .main-header {
        font-size: 2.5rem;
        color: #1f77b4;
        text-align: center;
        margin-bottom: 2rem;
    }
    
    .section-header {
        font-size: 1.5rem;
        color: #2c3e50;
        margin-top: 2rem;
        margin-bottom: 1rem;
        border-bottom: 2px solid #3498db;
        padding-bottom: 0.5rem;
    }
    
    .thinking-step {
        background-color: #e8f4fd;
        padding: 1rem;
        border-radius: 0.5rem;
        margin: 0.5rem 0;
        border-left: 4px solid #2ecc71;
        font-family: 'Courier New', monospace;
        font-size: 0.9rem;
        animation: fadeIn 0.5s ease-in;
    }
    
    .tavily-result {
        background-color: #fff3cd;
        padding: 1rem;
        border-radius: 0.5rem;
        margin: 0.5rem 0;
        border-left: 4px solid #f39c12;
        font-family: 'Courier New', monospace;
        font-size: 0.9rem;
        animation: slideIn 0.5s ease-in;
    }
    
    .firecrawl-result {
        background-color: #f8d7da;
        padding: 1rem;
        border-radius: 0.5rem;
        margin: 0.5rem 0;
        border-left: 4px solid #e74c3c;
        font-family: 'Courier New', monospace;
        font-size: 0.9rem;
        animation: slideIn 0.5s ease-in;
    }
    
    .final-result {
        background-color: #d4edda;
        padding: 1.5rem;
        border-radius: 0.5rem;
        margin: 1rem 0;
        border-left: 4px solid #28a745;
        animation: slideIn 0.8s ease-in;
    }
    
    .error-message {
        background-color: #f8d7da;
        padding: 1rem;
        border-radius: 0.5rem;
        margin: 0.5rem 0;
        border-left: 4px solid #dc3545;
        color: #721c24;
    }
    
    .process-step {
        background-color: #f8f9fa;
        padding: 1rem;
        border-radius: 0.5rem;
        margin: 1rem 0;
        border: 1px solid #dee2e6;
    }
    
    .api-status {
        padding: 0.5rem 1rem;
        border-radius: 0.25rem;
        margin: 0.5rem 0;
        font-weight: bold;
    }
    
    .api-available {
        background-color: #d4edda;
        color: #155724;
        border: 1px solid #c3e6cb;
    }
    
    .api-demo {
        background-color: #fff3cd;
        color: #856404;
        border: 1px solid #ffeaa7;
    }
    
    .influencer-card {
        background-color: #f8f9fa;
        padding: 1.5rem;
        border-radius: 0.5rem;
        margin: 1rem 0;
        border-left: 4px solid #6f42c1;
        box-shadow: 0 2px 4px rgba(0,0,0,0.1);
    }
    
    .real-time-log {
        max-height: 400px;
        overflow-y: auto;
        background-color: #2c3e50;
        color: #ecf0f1;
        padding: 1rem;
        border-radius: 0.5rem;
        font-family: 'Courier New', monospace;
        font-size: 0.8rem;
        margin: 1rem 0;
    }
    
    @keyframes fadeIn {
        from { opacity: 0; }
        to { opacity: 1; }
    }
    
    @keyframes slideIn {
        from { transform: translateX(-20px); opacity: 0; }
        to { transform: translateX(0); opacity: 1; }
    }
    
    .progress-container {
        background-color: #f8f9fa;
        padding: 1rem;
        border-radius: 0.5rem;
        margin: 1rem 0;
        border: 1px solid #dee2e6;
    }
    
    .step-indicator {
        display: inline-block;
        padding: 0.25rem 0.5rem;
        margin: 0.25rem;
        border-radius: 0.25rem;
        font-size: 0.8rem;
        font-weight: bold;
    }
    
    .step-pending {
        background-color: #e9ecef;
        color: #6c757d;
    }
    
    .step-active {
        background-color: #007bff;
        color: white;
        animation: pulse 1s infinite;
    }
    
    .step-complete {
        background-color: #28a745;
        color: white;
    }
    
    @keyframes pulse {
        0% { opacity: 1; }
        50% { opacity: 0.5; }
        100% { opacity: 1; }
    }
    </style>
    """, unsafe_allow_html=True)

def display_process_steps(current_step: str = "planning"):
    """Display the current process step indicators"""
    steps = [
        ("planning", "🧠 Planning"),
        ("thinking", "💭 Thinking"),
        ("tavily", "🌐 Tavily Search"),
        ("firecrawl", "🔍 Firecrawl Extract"),
        ("analysis", "📊 Analysis"),
        ("complete", "✅ Complete")
    ]
    
    step_html = '<div class="progress-container"><h4>Process Progress:</h4>'
    
    for step_id, step_name in steps:
        if step_id == current_step:
            css_class = "step-active"
        elif steps.index((step_id, step_name)) < steps.index((current_step, next(name for sid, name in steps if sid == current_step))):
            css_class = "step-complete"
        else:
            css_class = "step-pending"
        
        step_html += f'<span class="step-indicator {css_class}">{step_name}</span>'
    
    step_html += '</div>'
    st.markdown(step_html, unsafe_allow_html=True)

def create_real_time_log_container():
    """Create a container for real-time logging updates"""
    if 'log_messages' not in st.session_state:
        st.session_state.log_messages = []
    
    log_container = st.empty()
    return log_container

def update_real_time_log(container, update_data: Dict):
    """Update the real-time log display"""
    if 'log_messages' not in st.session_state:
        st.session_state.log_messages = []
    
    # Add new message
    timestamp = datetime.now().strftime("%H:%M:%S")
    update_type = update_data.get('type', 'info')
    content = update_data.get('content', '')
    
    st.session_state.log_messages.append({
        'timestamp': timestamp,
        'type': update_type,
        'content': content,
        'data': update_data.get('data')
    })
    
    # Keep only last 20 messages
    if len(st.session_state.log_messages) > 20:
        st.session_state.log_messages = st.session_state.log_messages[-20:]
    
    # Update display
    log_html = '<div class="real-time-log">'
    for msg in st.session_state.log_messages:
        log_html += f'<div>[{msg["timestamp"]}] <strong>{msg["type"].upper()}:</strong> {msg["content"]}</div>'
    log_html += '</div>'
    
    container.markdown(log_html, unsafe_allow_html=True)

def display_thinking_steps(updates: List[Dict]):
    """Display thinking steps in real-time"""
    st.markdown("### 🧠 BMasterAI Thinking Process")
    
    for update in updates:
        if update['type'] == 'thinking':
            st.markdown(f"""
            <div class="thinking-step">
                <strong>💭 Thinking:</strong> {update['content']}
                <br><small>⏰ {update.get('timestamp', '')}</small>
            </div>
            """, unsafe_allow_html=True)

def display_tavily_results(updates: List[Dict]):
    """Display Tavily search results"""
    tavily_updates = [u for u in updates if u['type'] in ['tavily_search', 'tavily_result']]
    
    if tavily_updates:
        st.markdown("### 🌐 Tavily Search Results")
        
        for update in tavily_updates:
            if update['type'] == 'tavily_search':
                st.markdown(f"""
                <div class="tavily-result">
                    <strong>🔍 Search:</strong> {update['content']}
                </div>
                """, unsafe_allow_html=True)
            elif update['type'] == 'tavily_result':
                data = update.get('data', [])
                st.markdown(f"""
                <div class="tavily-result">
                    <strong>📊 Results:</strong> {update['content']}
                    <br><small>Found {len(data) if isinstance(data, list) else 0} results</small>
                </div>
                """, unsafe_allow_html=True)
                
                # Show first few results
                if isinstance(data, list) and data:
                    with st.expander(f"View {len(data)} search results", expanded=False):
                        for i, result in enumerate(data[:3]):  # Show first 3
                            st.write(f"**{i+1}.** {result.get('title', 'No title')}")
                            st.write(f"URL: {result.get('url', 'No URL')}")
                            st.write(f"Content: {result.get('content', 'No content')[:200]}...")
                            st.write("---")

def display_firecrawl_results(updates: List[Dict]):
    """Display Firecrawl extraction results"""
    firecrawl_updates = [u for u in updates if u['type'] in ['firecrawl_crawl', 'firecrawl_result']]
    
    if firecrawl_updates:
        st.markdown("### 🔍 Firecrawl Email Extraction")
        
        for update in firecrawl_updates:
            if update['type'] == 'firecrawl_crawl':
                st.markdown(f"""
                <div class="firecrawl-result">
                    <strong>🕷️ Crawling:</strong> {update['content']}
                </div>
                """, unsafe_allow_html=True)
            elif update['type'] == 'firecrawl_result':
                data = update.get('data')
                st.markdown(f"""
                <div class="firecrawl-result">
                    <strong>📧 Extraction:</strong> {update['content']}
                </div>
                """, unsafe_allow_html=True)
                
                # Show extracted data
                if data and isinstance(data, dict):
                    emails = data.get('emails', [])
                    if emails:
                        with st.expander(f"📧 Found {len(emails)} email(s)", expanded=False):
                            for email in emails:
                                st.write(f"• {email}")
                            
                            extracted_info = data.get('extracted_info', {})
                            if extracted_info:
                                st.write("**Additional Info:**")
                                for key, value in extracted_info.items():
                                    st.write(f"• {key}: {value}")

def main():
    """Main Streamlit application with real-time BMasterAI logging"""
    st.set_page_config(
        page_title="Gemini + BMasterAI + Tavily + Firecrawl",
        page_icon="🧠",
        layout="wide",
        initial_sidebar_state="expanded"
    )
    
    load_css()
    
    # Header
    st.markdown('<div class="main-header">🧠 Gemini + BMasterAI + Tavily + Firecrawl</div>', unsafe_allow_html=True)
    st.markdown("**AI Podcast Influencer Research with Real-Time Reasoning Transparency**")
    st.markdown("---")
    
    # Check BMasterAI availability
    if not BMASTERAI_AVAILABLE:
        st.error("❌ BMasterAI library is required for this application. Please install it first.")
        st.code("pip install bmasterai")
        st.stop()
    
    # Sidebar for configuration
    with st.sidebar:
        st.header("🔑 API Configuration")
        
        # Gemini API Key
        env_gemini_key = os.getenv("GEMINI_API_KEY")
        if env_gemini_key:
            st.success("✅ Gemini API key loaded from environment")
            gemini_api_key = env_gemini_key
        else:
            st.warning("⚠️ No Gemini API key found in environment")
            gemini_api_key = st.text_input(
                "Enter your Gemini API Key:",
                type="password",
                help="Get your API key from Google AI Studio: https://makersuite.google.com/app/apikey"
            )
        
        # Tavily API Key
        env_tavily_key = os.getenv("TAVILY_API_KEY")
        if env_tavily_key:
            st.markdown('<div class="api-status api-available">✅ Tavily API key loaded</div>', unsafe_allow_html=True)
            tavily_api_key = env_tavily_key
        else:
            st.markdown('<div class="api-status api-demo">ℹ️ No Tavily API key found</div>', unsafe_allow_html=True)
            tavily_api_key = st.text_input(
                "Enter your Tavily API Key (optional):",
                type="password",
                help="Get your API key from Tavily: https://tavily.com/"
            )
        
        # Firecrawl API Key
        env_firecrawl_key = os.getenv("FIRECRAWL_API_KEY")
        if env_firecrawl_key:
            st.markdown('<div class="api-status api-available">✅ Firecrawl API key loaded</div>', unsafe_allow_html=True)
            firecrawl_api_key = env_firecrawl_key
        else:
            st.markdown('<div class="api-status api-demo">ℹ️ No Firecrawl API key found</div>', unsafe_allow_html=True)
            firecrawl_api_key = st.text_input(
                "Enter your Firecrawl API Key (optional):",
                type="password",
                help="Get your API key from Firecrawl: https://firecrawl.dev/"
            )
        
        if not gemini_api_key:
            st.error("Please provide a Gemini API key to continue")
            st.stop()
        
        st.markdown("---")
        
        # BMasterAI Configuration
        st.header("🔧 BMasterAI Setup")
        
        if st.button("Initialize BMasterAI Logging"):
            try:
                configure_logging(
                    log_level=LogLevel.DEBUG,
                    enable_console=True,
                    enable_reasoning_logs=True,
                    reasoning_log_file="gemini_tavily_firecrawl_reasoning.jsonl"
                )
                st.success("✅ BMasterAI logging configured!")
            except Exception as e:
                st.error(f"❌ Error configuring BMasterAI: {str(e)}")
        
        st.markdown("---")
        st.header("📋 Process Overview")
        st.markdown("""
        **Real-Time Process Flow:**
        1. 🧠 **BMasterAI Thinking** - See Gemini's reasoning
        2. 🌐 **Tavily Search** - Find podcast websites
        3. 🔍 **Firecrawl Extract** - Get email addresses
        4. 📊 **Analysis** - Structure final results
        5. ✅ **Complete** - Present verified contacts
        """)
    
    # Initialize session state for real-time updates
    if 'updates' not in st.session_state:
        st.session_state.updates = []
    
    if 'agent' not in st.session_state:
        try:
            st.session_state.agent = GeminiReasoningAgent(
                "streamlit-gemini-tavily-firecrawl-agent", 
                gemini_api_key, 
                tavily_api_key if tavily_api_key else None,
                firecrawl_api_key if firecrawl_api_key else None
            )
        except Exception as e:
            st.error(f"Failed to initialize Gemini agent: {str(e)}")
            st.stop()
    
    # Main content
    st.markdown('<div class="section-header">🎙️ AI Podcast Influencer Search with Real-Time Reasoning</div>', unsafe_allow_html=True)
    st.write("Watch Gemini's complete thinking process as it searches for AI podcast influencers using Tavily + Firecrawl APIs.")
    
    # Search parameters
    col1, col2 = st.columns(2)
    
    with col1:
        target_count = st.number_input(
            "Number of influencers to find:",
            min_value=1,
            max_value=20,
            value=5,
            help="How many AI podcast influencers to search for"
        )
    
    with col2:
        focus_area = st.selectbox(
            "Focus area:",
            [
                "general AI",
                "machine learning and AI applications", 
                "AI research and academia",
                "AI in business and enterprise",
                "AI ethics and society",
                "computer vision and NLP",
                "robotics and autonomous systems"
            ],
            help="Specific area of AI to focus the search on"
        )
    
    # API status display
    api_status_col1, api_status_col2, api_status_col3 = st.columns(3)
    
    with api_status_col1:
        st.markdown('<div class="api-status api-available">🧠 Gemini: Ready</div>', unsafe_allow_html=True)
    
    with api_status_col2:
        if tavily_api_key:
            st.markdown('<div class="api-status api-available">🌐 Tavily: Live Search</div>', unsafe_allow_html=True)
        else:
            st.markdown('<div class="api-status api-demo">🌐 Tavily: Demo Mode</div>', unsafe_allow_html=True)
    
    with api_status_col3:
        if firecrawl_api_key:
            st.markdown('<div class="api-status api-available">🔍 Firecrawl: Live Extract</div>', unsafe_allow_html=True)
        else:
            st.markdown('<div class="api-status api-demo">🔍 Firecrawl: Demo Mode</div>', unsafe_allow_html=True)
    
    # Search button
    if st.button("🚀 Start AI Podcast Influencer Search", type="primary"):
        # Clear previous updates
        st.session_state.updates = []
        
        # Create containers for real-time updates
        progress_container = st.empty()
        thinking_container = st.empty()
        tavily_container = st.empty()
        firecrawl_container = st.empty()
        results_container = st.empty()
        
        # Set up real-time callback
        def update_callback(update_data):
            st.session_state.updates.append(update_data)
            
            # Update progress indicator
            current_step = "planning"
            if any(u['type'] == 'thinking' for u in st.session_state.updates):
                current_step = "thinking"
            if any(u['type'] in ['tavily_search', 'tavily_result'] for u in st.session_state.updates):
                current_step = "tavily"
            if any(u['type'] in ['firecrawl_crawl', 'firecrawl_result'] for u in st.session_state.updates):
                current_step = "firecrawl"
            if any(u['type'] == 'conclusion' for u in st.session_state.updates):
                current_step = "complete"
            
            with progress_container.container():
                display_process_steps(current_step)
            
            # Update thinking steps
            thinking_updates = [u for u in st.session_state.updates if u['type'] == 'thinking']
            if thinking_updates:
                with thinking_container.container():
                    display_thinking_steps(thinking_updates)
            
            # Update Tavily results
            tavily_updates = [u for u in st.session_state.updates if u['type'] in ['tavily_search', 'tavily_result']]
            if tavily_updates:
                with tavily_container.container():
                    display_tavily_results(tavily_updates)
            
            # Update Firecrawl results
            firecrawl_updates = [u for u in st.session_state.updates if u['type'] in ['firecrawl_crawl', 'firecrawl_result']]
            if firecrawl_updates:
                with firecrawl_container.container():
                    display_firecrawl_results(firecrawl_updates)
            
            # Update final results
            final_updates = [u for u in st.session_state.updates if u['type'] in ['final_results', 'conclusion']]
            if final_updates:
                with results_container.container():
                    st.markdown("### ✅ Final Results")
                    for update in final_updates:
                        if update['type'] == 'final_results':
                            data = update.get('data', [])
                            st.markdown(f"""
                            <div class="final-result">
                                <h4>🎯 {update['content']}</h4>
                                <p>Found {len(data) if isinstance(data, list) else 0} verified influencer profiles</p>
                            </div>
                            """, unsafe_allow_html=True)
                            
                            # Display influencer cards
                            if isinstance(data, list):
                                for i, influencer in enumerate(data, 1):
                                    st.markdown(f"""
                                    <div class="influencer-card">
                                        <h4>{i}. {influencer.get('name', 'Unknown')}</h4>
                                        <p><strong>Podcast:</strong> {influencer.get('podcast', 'N/A')}</p>
                                        <p><strong>Email:</strong> {influencer.get('email', 'Not found')}</p>
                                        <p><strong>Website:</strong> {influencer.get('website', 'N/A')}</p>
                                        <p><strong>Focus:</strong> {influencer.get('focus_area', 'General AI')}</p>
                                        <p><strong>Guest Friendly:</strong> {influencer.get('guest_friendly', 'Unknown')}</p>
                                        <p><strong>Contact Confidence:</strong> {influencer.get('contact_confidence', 'N/A')}/10</p>
                                        <p><strong>Notes:</strong> {influencer.get('notes', 'No additional notes')}</p>
                                    </div>
                                    """, unsafe_allow_html=True)
                        
                        elif update['type'] == 'conclusion':
                            st.markdown(f"""
                            <div class="final-result">
                                <h4>🎯 Process Complete</h4>
                                <p>{update['content']}</p>
                            </div>
                            """, unsafe_allow_html=True)
            
            # Handle errors
            error_updates = [u for u in st.session_state.updates if u['type'] == 'error']
            if error_updates:
                for error_update in error_updates:
                    st.markdown(f"""
                    <div class="error-message">
                        <strong>❌ Error:</strong> {error_update['content']}
                    </div>
                    """, unsafe_allow_html=True)
        
        # Set callback and run search
        st.session_state.agent.set_update_callback(update_callback)
        
        with st.spinner("🧠 Starting AI podcast influencer search with real-time reasoning..."):
            try:
                result = st.session_state.agent.search_ai_podcast_influencers_with_firecrawl(
                    target_count=target_count,
                    focus_area=focus_area
                )
                
                st.success("✅ Search completed with full BMasterAI + Tavily + Firecrawl integration!")
                
            except Exception as e:
                st.error(f"❌ Error during search: {str(e)}")
    
    # Additional tabs for other features
    st.markdown("---")
    
    tab1, tab2, tab3, tab4 = st.tabs([
        "🎭 Sentiment Analysis", 
        "🧩 Problem Solving", 
        "❓ Query Processing", 
        "📊 Logs & Analytics"
    ])
    
    with tab1:
        st.markdown("### 🎭 Sentiment Analysis with Real-Time Thinking")
        
        text_input = st.text_area(
            "Enter text to analyze:",
            placeholder="e.g., I'm excited about the potential of AI podcasts to educate and inspire!",
            height=100
        )
        
        if st.button("Analyze Sentiment", key="sentiment"):
            if text_input.strip():
                sentiment_updates = []
                
                def sentiment_callback(update_data):
                    sentiment_updates.append(update_data)
                    if update_data['type'] == 'thinking':
                        st.markdown(f"""
                        <div class="thinking-step">
                            <strong>💭 Thinking:</strong> {update_data['content']}
                        </div>
                        """, unsafe_allow_html=True)
                
                st.session_state.agent.set_update_callback(sentiment_callback)
                
                with st.spinner("Analyzing sentiment with real-time reasoning..."):
                    try:
                        result = st.session_state.agent.analyze_sentiment_with_context_manager(text_input)
                        st.success("✅ Sentiment analysis completed!")
                        st.write("**Result:**", result)
                    except Exception as e:
                        st.error(f"Error during analysis: {str(e)}")
            else:
                st.warning("Please enter some text to analyze.")
    
    with tab2:
        st.markdown("### 🧩 Problem Solving with Chain of Thought")
        
        problem_input = st.text_area(
            "Enter a problem to solve:",
            placeholder="e.g., How can I effectively pitch myself as a guest on AI podcasts?",
            height=100
        )
        
        if st.button("Solve Problem", key="problem"):
            if problem_input.strip():
                problem_updates = []
                
                def problem_callback(update_data):
                    problem_updates.append(update_data)
                    if update_data['type'] == 'thinking':
                        st.markdown(f"""
                        <div class="thinking-step">
                            <strong>💭 Thinking:</strong> {update_data['content']}
                        </div>
                        """, unsafe_allow_html=True)
                
                st.session_state.agent.set_update_callback(problem_callback)
                
                with st.spinner("Solving problem with chain of thought reasoning..."):
                    try:
                        result = st.session_state.agent.analyze_with_chain_of_thought(problem_input)
                        st.success("✅ Problem solved!")
                        st.write("**Solution:**", result)
                    except Exception as e:
                        st.error(f"Error during problem solving: {str(e)}")
            else:
                st.warning("Please enter a problem to solve.")
    
    with tab3:
        st.markdown("### ❓ Query Processing with Real-Time Logging")
        
        query_input = st.text_area(
            "Enter your query:",
            placeholder="e.g., What are the best strategies for getting booked on AI podcasts?",
            height=100
        )
        
        if st.button("Process Query", key="query"):
            if query_input.strip():
                query_updates = []
                
                def query_callback(update_data):
                    query_updates.append(update_data)
                    if update_data['type'] == 'thinking':
                        st.markdown(f"""
                        <div class="thinking-step">
                            <strong>💭 Thinking:</strong> {update_data['content']}
                        </div>
                        """, unsafe_allow_html=True)
                
                st.session_state.agent.set_update_callback(query_callback)
                
                with st.spinner("Processing query with real-time reasoning..."):
                    try:
                        result = st.session_state.agent.process_with_decorator(
                            st.session_state.agent.agent_id, 
                            query_input
                        )
                        st.success("✅ Query processed!")
                        st.write("**Response:**", result)
                    except Exception as e:
                        st.error(f"Error during query processing: {str(e)}")
            else:
                st.warning("Please enter a query to process.")
    
    with tab4:
        st.markdown("### 📊 BMasterAI Logs & Analytics")
        
        # Display recent logs
        if st.button("Refresh Logs"):
            st.rerun()
        
        # Show agent statistics
        if st.button("Show Agent Statistics"):
            try:
                stats = st.session_state.agent.get_agent_stats()
                
                st.markdown("### 📈 Agent Statistics")
                col1, col2 = st.columns(2)
                
                with col1:
                    st.metric("Total Events", stats.get('total_events', 0))
                
                with col2:
                    event_types = stats.get('event_types', {})
                    if event_types:
                        st.write("**Event Types:**")
                        for event_type, count in event_types.items():
                            st.write(f"• {event_type}: {count}")
                
            except Exception as e:
                st.error(f"Error getting statistics: {str(e)}")
        
        # Export functionality
        st.markdown("---")
        st.markdown("### 📥 Export BMasterAI Logs")
        
        col1, col2 = st.columns(2)
        
        with col1:
            if st.button("Export as JSON"):
                try:
                    json_logs = st.session_state.agent.export_reasoning_logs("json")
                    st.download_button(
                        label="Download JSON Logs",
                        data=json_logs,
                        file_name=f"bmasterai_tavily_firecrawl_logs_{datetime.now().strftime('%Y%m%d_%H%M%S')}.json",
                        mime="application/json"
                    )
                except Exception as e:
                    st.error(f"Error exporting JSON: {str(e)}")
        
        with col2:
            if st.button("Export as Markdown"):
                try:
                    markdown_logs = st.session_state.agent.export_reasoning_logs("markdown")
                    st.download_button(
                        label="Download Markdown Logs",
                        data=markdown_logs,
                        file_name=f"bmasterai_tavily_firecrawl_logs_{datetime.now().strftime('%Y%m%d_%H%M%S')}.md",
                        mime="text/markdown"
                    )
                except Exception as e:
                    st.error(f"Error exporting Markdown: {str(e)}")
    
    # Footer
    st.markdown("---")
    st.markdown("""
    <div style="text-align: center; color: #7f8c8d; margin-top: 2rem;">
        <p>🧠 Powered by Gemini 2.5 Pro + BMasterAI + Tavily + Firecrawl | Built with Streamlit</p>
        <p><small>Real-time reasoning transparency for complex AI tasks with multi-API integrations.</small></p>
    </div>
    """, unsafe_allow_html=True)

if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.cloud.aiplatform_v1beta1.types import model_monitor as gca_model_monitor
from google.cloud.aiplatform_v1beta1.types import model_monitoring_alert
from google.cloud.aiplatform_v1beta1.types import (
    model_monitoring_job as gca_model_monitoring_job,
)
from google.cloud.aiplatform_v1beta1.types import model_monitoring_stats
from google.cloud.aiplatform_v1beta1.types import operation
from google.protobuf import field_mask_pb2  # type: ignore
from google.type import interval_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1beta1",
    manifest={
        "CreateModelMonitorRequest",
        "CreateModelMonitorOperationMetadata",
        "UpdateModelMonitorRequest",
        "UpdateModelMonitorOperationMetadata",
        "GetModelMonitorRequest",
        "ListModelMonitorsRequest",
        "ListModelMonitorsResponse",
        "DeleteModelMonitorRequest",
        "CreateModelMonitoringJobRequest",
        "GetModelMonitoringJobRequest",
        "ListModelMonitoringJobsRequest",
        "ListModelMonitoringJobsResponse",
        "DeleteModelMonitoringJobRequest",
        "SearchModelMonitoringStatsRequest",
        "SearchModelMonitoringStatsResponse",
        "SearchModelMonitoringAlertsRequest",
        "SearchModelMonitoringAlertsResponse",
    },
)


class CreateModelMonitorRequest(proto.Message):
    r"""Request message for
    [ModelMonitoringService.CreateModelMonitor][google.cloud.aiplatform.v1beta1.ModelMonitoringService.CreateModelMonitor].

    Attributes:
        parent (str):
            Required. The resource name of the Location to create the
            ModelMonitor in. Format:
            ``projects/{project}/locations/{location}``
        model_monitor (google.cloud.aiplatform_v1beta1.types.ModelMonitor):
            Required. The ModelMonitor to create.
        model_monitor_id (str):
            Optional. The ID to use for the Model Monitor, which will
            become the final component of the model monitor resource
            name.

            The maximum length is 63 characters, and valid characters
            are ``/^[a-z]([a-z0-9-]{0,61}[a-z0-9])?$/``.
    """

    parent: str = proto.Field(
        proto.STRING,
        number=1,
    )
    model_monitor: gca_model_monitor.ModelMonitor = proto.Field(
        proto.MESSAGE,
        number=2,
        message=gca_model_monitor.ModelMonitor,
    )
    model_monitor_id: str = proto.Field(
        proto.STRING,
        number=3,
    )


class CreateModelMonitorOperationMetadata(proto.Message):
    r"""Runtime operation information for
    [ModelMonitoringService.CreateModelMonitor][google.cloud.aiplatform.v1beta1.ModelMonitoringService.CreateModelMonitor].

    Attributes:
        generic_metadata (google.cloud.aiplatform_v1beta1.types.GenericOperationMetadata):
            The operation generic information.
    """

    generic_metadata: operation.GenericOperationMetadata = proto.Field(
        proto.MESSAGE,
        number=1,
        message=operation.GenericOperationMetadata,
    )


class UpdateModelMonitorRequest(proto.Message):
    r"""Request message for
    [ModelMonitoringService.UpdateModelMonitor][google.cloud.aiplatform.v1beta1.ModelMonitoringService.UpdateModelMonitor].

    Attributes:
        model_monitor (google.cloud.aiplatform_v1beta1.types.ModelMonitor):
            Required. The model monitoring configuration
            which replaces the resource on the server.
        update_mask (google.protobuf.field_mask_pb2.FieldMask):
            Required. Mask specifying which fields to
            update.
    """

    model_monitor: gca_model_monitor.ModelMonitor = proto.Field(
        proto.MESSAGE,
        number=1,
        message=gca_model_monitor.ModelMonitor,
    )
    update_mask: field_mask_pb2.FieldMask = proto.Field(
        proto.MESSAGE,
        number=2,
        message=field_mask_pb2.FieldMask,
    )


class UpdateModelMonitorOperationMetadata(proto.Message):
    r"""Runtime operation information for
    [ModelMonitoringService.UpdateModelMonitor][google.cloud.aiplatform.v1beta1.ModelMonitoringService.UpdateModelMonitor].

    Attributes:
        generic_metadata (google.cloud.aiplatform_v1beta1.types.GenericOperationMetadata):
            The operation generic information.
    """

    generic_metadata: operation.GenericOperationMetadata = proto.Field(
        proto.MESSAGE,
        number=1,
        message=operation.GenericOperationMetadata,
    )


class GetModelMonitorRequest(proto.Message):
    r"""Request message for
    [ModelMonitoringService.GetModelMonitor][google.cloud.aiplatform.v1beta1.ModelMonitoringService.GetModelMonitor].

    Attributes:
        name (str):
            Required. The name of the ModelMonitor resource. Format:
            ``projects/{project}/locations/{location}/modelMonitors/{model_monitor}``
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )


class ListModelMonitorsRequest(proto.Message):
    r"""Request message for
    [ModelMonitoringService.ListModelMonitors][google.cloud.aiplatform.v1beta1.ModelMonitoringService.ListModelMonitors].

    Attributes:
        parent (str):
            Required. The resource name of the Location to list the
            ModelMonitors from. Format:
            ``projects/{project}/locations/{location}``
        filter (str):
            The standard list filter. More detail in
            `AIP-160 <https://google.aip.dev/160>`__.
        page_size (int):
            The standard list page size.
        page_token (str):
            The standard list page token.
        read_mask (google.protobuf.field_mask_pb2.FieldMask):
            Mask specifying which fields to read.
    """

    parent: str = proto.Field(
        proto.STRING,
        number=1,
    )
    filter: str = proto.Field(
        proto.STRING,
        number=2,
    )
    page_size: int = proto.Field(
        proto.INT32,
        number=3,
    )
    page_token: str = proto.Field(
        proto.STRING,
        number=4,
    )
    read_mask: field_mask_pb2.FieldMask = proto.Field(
        proto.MESSAGE,
        number=5,
        message=field_mask_pb2.FieldMask,
    )


class ListModelMonitorsResponse(proto.Message):
    r"""Response message for
    [ModelMonitoringService.ListModelMonitors][google.cloud.aiplatform.v1beta1.ModelMonitoringService.ListModelMonitors]

    Attributes:
        model_monitors (MutableSequence[google.cloud.aiplatform_v1beta1.types.ModelMonitor]):
            List of ModelMonitor in the requested page.
        next_page_token (str):
            A token to retrieve the next page of results. Pass to
            [ListModelMonitorsRequest.page_token][google.cloud.aiplatform.v1beta1.ListModelMonitorsRequest.page_token]
            to obtain that page.
    """

    @property
    def raw_page(self):
        return self

    model_monitors: MutableSequence[gca_model_monitor.ModelMonitor] = (
        proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message=gca_model_monitor.ModelMonitor,
        )
    )
    next_page_token: str = proto.Field(
        proto.STRING,
        number=2,
    )


class DeleteModelMonitorRequest(proto.Message):
    r"""Request message for
    [ModelMonitoringService.DeleteModelMonitor][google.cloud.aiplatform.v1beta1.ModelMonitoringService.DeleteModelMonitor].

    Attributes:
        name (str):
            Required. The name of the ModelMonitor resource to be
            deleted. Format:
            ``projects/{project}/locations/{location}/modelMonitords/{model_monitor}``
        force (bool):
            Optional. Force delete the model monitor with
            schedules.
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    force: bool = proto.Field(
        proto.BOOL,
        number=2,
    )


class CreateModelMonitoringJobRequest(proto.Message):
    r"""Request message for
    [ModelMonitoringService.CreateModelMonitoringJob][google.cloud.aiplatform.v1beta1.ModelMonitoringService.CreateModelMonitoringJob].

    Attributes:
        parent (str):
            Required. The parent of the ModelMonitoringJob. Format:
            ``projects/{project}/locations/{location}/modelMoniitors/{model_monitor}``
        model_monitoring_job (google.cloud.aiplatform_v1beta1.types.ModelMonitoringJob):
            Required. The ModelMonitoringJob to create
        model_monitoring_job_id (str):
            Optional. The ID to use for the Model Monitoring Job, which
            will become the final component of the model monitoring job
            resource name.

            The maximum length is 63 characters, and valid characters
            are ``/^[a-z]([a-z0-9-]{0,61}[a-z0-9])?$/``.
    """

    parent: str = proto.Field(
        proto.STRING,
        number=1,
    )
    model_monitoring_job: gca_model_monitoring_job.ModelMonitoringJob = proto.Field(
        proto.MESSAGE,
        number=2,
        message=gca_model_monitoring_job.ModelMonitoringJob,
    )
    model_monitoring_job_id: str = proto.Field(
        proto.STRING,
        number=3,
    )


class GetModelMonitoringJobRequest(proto.Message):
    r"""Request message for
    [ModelMonitoringService.GetModelMonitoringJob][google.cloud.aiplatform.v1beta1.ModelMonitoringService.GetModelMonitoringJob].

    Attributes:
        name (str):
            Required. The resource name of the ModelMonitoringJob.
            Format:
            ``projects/{project}/locations/{location}/modelMonitors/{model_monitor}/modelMonitoringJobs/{model_monitoring_job}``
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )


class ListModelMonitoringJobsRequest(proto.Message):
    r"""Request message for
    [ModelMonitoringService.ListModelMonitoringJobs][google.cloud.aiplatform.v1beta1.ModelMonitoringService.ListModelMonitoringJobs].

    Attributes:
        parent (str):
            Required. The parent of the ModelMonitoringJob. Format:
            ``projects/{project}/locations/{location}/modelMonitors/{model_monitor}``
        filter (str):
            The standard list filter. More detail in
            `AIP-160 <https://google.aip.dev/160>`__.
        page_size (int):
            The standard list page size.
        page_token (str):
            The standard list page token.
        read_mask (google.protobuf.field_mask_pb2.FieldMask):
            Mask specifying which fields to read
    """

    parent: str = proto.Field(
        proto.STRING,
        number=1,
    )
    filter: str = proto.Field(
        proto.STRING,
        number=2,
    )
    page_size: int = proto.Field(
        proto.INT32,
        number=3,
    )
    page_token: str = proto.Field(
        proto.STRING,
        number=4,
    )
    read_mask: field_mask_pb2.FieldMask = proto.Field(
        proto.MESSAGE,
        number=5,
        message=field_mask_pb2.FieldMask,
    )


class ListModelMonitoringJobsResponse(proto.Message):
    r"""Response message for
    [ModelMonitoringService.ListModelMonitoringJobs][google.cloud.aiplatform.v1beta1.ModelMonitoringService.ListModelMonitoringJobs].

    Attributes:
        model_monitoring_jobs (MutableSequence[google.cloud.aiplatform_v1beta1.types.ModelMonitoringJob]):
            A list of ModelMonitoringJobs that matches
            the specified filter in the request.
        next_page_token (str):
            The standard List next-page token.
    """

    @property
    def raw_page(self):
        return self

    model_monitoring_jobs: MutableSequence[
        gca_model_monitoring_job.ModelMonitoringJob
    ] = proto.RepeatedField(
        proto.MESSAGE,
        number=1,
        message=gca_model_monitoring_job.ModelMonitoringJob,
    )
    next_page_token: str = proto.Field(
        proto.STRING,
        number=2,
    )


class DeleteModelMonitoringJobRequest(proto.Message):
    r"""Request message for
    [ModelMonitoringService.DeleteModelMonitoringJob][google.cloud.aiplatform.v1beta1.ModelMonitoringService.DeleteModelMonitoringJob].

    Attributes:
        name (str):
            Required. The resource name of the model monitoring job to
            delete. Format:
            ``projects/{project}/locations/{location}/modelMonitors/{model_monitor}/modelMonitoringJobs/{model_monitoring_job}``
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )


class SearchModelMonitoringStatsRequest(proto.Message):
    r"""Request message for
    [ModelMonitoringService.SearchModelMonitoringStats][google.cloud.aiplatform.v1beta1.ModelMonitoringService.SearchModelMonitoringStats].

    Attributes:
        model_monitor (str):
            Required. ModelMonitor resource name. Format:
            ``projects/{project}/locations/{location}/modelMonitors/{model_monitor}``
        stats_filter (google.cloud.aiplatform_v1beta1.types.SearchModelMonitoringStatsFilter):
            Filter for search different stats.
        time_interval (google.type.interval_pb2.Interval):
            The time interval for which results should be
            returned.
        page_size (int):
            The standard list page size.
        page_token (str):
            A page token received from a previous
            [ModelMonitoringService.SearchModelMonitoringStats][google.cloud.aiplatform.v1beta1.ModelMonitoringService.SearchModelMonitoringStats]
            call.
    """

    model_monitor: str = proto.Field(
        proto.STRING,
        number=1,
    )
    stats_filter: model_monitoring_stats.SearchModelMonitoringStatsFilter = proto.Field(
        proto.MESSAGE,
        number=2,
        message=model_monitoring_stats.SearchModelMonitoringStatsFilter,
    )
    time_interval: interval_pb2.Interval = proto.Field(
        proto.MESSAGE,
        number=3,
        message=interval_pb2.Interval,
    )
    page_size: int = proto.Field(
        proto.INT32,
        number=4,
    )
    page_token: str = proto.Field(
        proto.STRING,
        number=5,
    )


class SearchModelMonitoringStatsResponse(proto.Message):
    r"""Response message for
    [ModelMonitoringService.SearchModelMonitoringStats][google.cloud.aiplatform.v1beta1.ModelMonitoringService.SearchModelMonitoringStats].

    Attributes:
        monitoring_stats (MutableSequence[google.cloud.aiplatform_v1beta1.types.ModelMonitoringStats]):
            Stats retrieved for requested objectives.
        next_page_token (str):
            The page token that can be used by the next
            [ModelMonitoringService.SearchModelMonitoringStats][google.cloud.aiplatform.v1beta1.ModelMonitoringService.SearchModelMonitoringStats]
            call.
    """

    @property
    def raw_page(self):
        return self

    monitoring_stats: MutableSequence[model_monitoring_stats.ModelMonitoringStats] = (
        proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message=model_monitoring_stats.ModelMonitoringStats,
        )
    )
    next_page_token: str = proto.Field(
        proto.STRING,
        number=2,
    )


class SearchModelMonitoringAlertsRequest(proto.Message):
    r"""Request message for
    [ModelMonitoringService.SearchModelMonitoringAlerts][google.cloud.aiplatform.v1beta1.ModelMonitoringService.SearchModelMonitoringAlerts].

    Attributes:
        model_monitor (str):
            Required. ModelMonitor resource name. Format:
            ``projects/{project}/locations/{location}/modelMonitors/{model_monitor}``
        model_monitoring_job (str):
            If non-empty, returns the alerts of this
            model monitoring job.
        alert_time_interval (google.type.interval_pb2.Interval):
            If non-empty, returns the alerts in this time
            interval.
        stats_name (str):
            If non-empty, returns the alerts of this stats_name.
        objective_type (str):
            If non-empty, returns the alerts of this objective type.
            Supported monitoring objectives: ``raw-feature-drift``
            ``prediction-output-drift`` ``feature-attribution``
        page_size (int):
            The standard list page size.
        page_token (str):
            A page token received from a previous
            [ModelMonitoringService.SearchModelMonitoringAlerts][google.cloud.aiplatform.v1beta1.ModelMonitoringService.SearchModelMonitoringAlerts]
            call.
    """

    model_monitor: str = proto.Field(
        proto.STRING,
        number=1,
    )
    model_monitoring_job: str = proto.Field(
        proto.STRING,
        number=2,
    )
    alert_time_interval: interval_pb2.Interval = proto.Field(
        proto.MESSAGE,
        number=3,
        message=interval_pb2.Interval,
    )
    stats_name: str = proto.Field(
        proto.STRING,
        number=4,
    )
    objective_type: str = proto.Field(
        proto.STRING,
        number=5,
    )
    page_size: int = proto.Field(
        proto.INT32,
        number=6,
    )
    page_token: str = proto.Field(
        proto.STRING,
        number=7,
    )


class SearchModelMonitoringAlertsResponse(proto.Message):
    r"""Response message for
    [ModelMonitoringService.SearchModelMonitoringAlerts][google.cloud.aiplatform.v1beta1.ModelMonitoringService.SearchModelMonitoringAlerts].

    Attributes:
        model_monitoring_alerts (MutableSequence[google.cloud.aiplatform_v1beta1.types.ModelMonitoringAlert]):
            Alerts retrieved for the requested
            objectives. Sorted by alert time descendingly.
        total_number_alerts (int):
            The total number of alerts retrieved by the
            requested objectives.
        next_page_token (str):
            The page token that can be used by the next
            [ModelMonitoringService.SearchModelMonitoringAlerts][google.cloud.aiplatform.v1beta1.ModelMonitoringService.SearchModelMonitoringAlerts]
            call.
    """

    @property
    def raw_page(self):
        return self

    model_monitoring_alerts: MutableSequence[
        model_monitoring_alert.ModelMonitoringAlert
    ] = proto.RepeatedField(
        proto.MESSAGE,
        number=1,
        message=model_monitoring_alert.ModelMonitoringAlert,
    )
    total_number_alerts: int = proto.Field(
        proto.INT64,
        number=2,
    )
    next_page_token: str = proto.Field(
        proto.STRING,
        number=3,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# Copyright 2016 Google LLC All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrap representation of Spanner keys / ranges."""

from google.cloud.spanner_v1 import KeyRangePB
from google.cloud.spanner_v1 import KeySetPB

from google.cloud.spanner_v1._helpers import _make_list_value_pb
from google.cloud.spanner_v1._helpers import _make_list_value_pbs


class KeyRange(object):
    """Identify range of table rows via start / end points.

    Specify either a `start_open` or `start_closed` key, or defaults to
    `start_closed = []`.  Specify either an `end_open` or `end_closed` key,
    or defaults to `end_closed = []`.  However, at least one key has to be
    specified.  If no keys are specified, ValueError is raised.

    :type start_open: list of scalars
    :param start_open: keys identifying start of range (this key excluded)

    :type start_closed: list of scalars
    :param start_closed: keys identifying start of range (this key included)

    :type end_open: list of scalars
    :param end_open: keys identifying end of range (this key excluded)

    :type end_closed: list of scalars
    :param end_closed: keys identifying end of range (this key included)

    :raises ValueError: if no keys are specified
    """

    def __init__(
        self, start_open=None, start_closed=None, end_open=None, end_closed=None
    ):
        if not any([start_open, start_closed, end_open, end_closed]):
            raise ValueError("Must specify at least a start or end row.")

        if start_open and start_closed:
            raise ValueError("Specify one of 'start_open' / 'start_closed'.")
        elif start_open is None and start_closed is None:
            start_closed = []

        if end_open and end_closed:
            raise ValueError("Specify one of 'end_open' / 'end_closed'.")
        elif end_open is None and end_closed is None:
            end_closed = []

        self.start_open = start_open
        self.start_closed = start_closed
        self.end_open = end_open
        self.end_closed = end_closed

    def _to_pb(self):
        """Construct a KeyRange protobuf.

        :rtype: :class:`~google.cloud.spanner_v1.types.KeyRange`
        :returns: protobuf corresponding to this instance.
        """
        kwargs = {}

        if self.start_open is not None:
            kwargs["start_open"] = _make_list_value_pb(self.start_open)

        if self.start_closed is not None:
            kwargs["start_closed"] = _make_list_value_pb(self.start_closed)

        if self.end_open is not None:
            kwargs["end_open"] = _make_list_value_pb(self.end_open)

        if self.end_closed is not None:
            kwargs["end_closed"] = _make_list_value_pb(self.end_closed)

        return KeyRangePB(**kwargs)

    def _to_dict(self):
        """Return the state of the keyrange as a dict.

        :rtype: dict
        :returns: state of this instance.
        """
        mapping = {}

        if self.start_open:
            mapping["start_open"] = self.start_open

        if self.start_closed:
            mapping["start_closed"] = self.start_closed

        if self.end_open:
            mapping["end_open"] = self.end_open

        if self.end_closed:
            mapping["end_closed"] = self.end_closed

        return mapping

    def __eq__(self, other):
        """Compare by serialized state."""
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self._to_dict() == other._to_dict()


class KeySet(object):
    """Identify table rows via keys / ranges.

    :type keys: list of list of scalars
    :param keys: keys identifying individual rows within a table.

    :type ranges: list of :class:`KeyRange`
    :param ranges: ranges identifying rows within a table.

    :type all_: boolean
    :param all_: if True, identify all rows within a table
    """

    def __init__(self, keys=(), ranges=(), all_=False):
        if all_ and (keys or ranges):
            raise ValueError("'all_' is exclusive of 'keys' / 'ranges'.")
        self.keys = list(keys)
        self.ranges = list(ranges)
        self.all_ = all_

    def _to_pb(self):
        """Construct a KeySet protobuf.

        :rtype: :class:`~google.cloud.spanner_v1.types.KeySet`
        :returns: protobuf corresponding to this instance.
        """
        if self.all_:
            return KeySetPB(all_=True)
        kwargs = {}

        if self.keys:
            kwargs["keys"] = _make_list_value_pbs(self.keys)

        if self.ranges:
            kwargs["ranges"] = [krange._to_pb() for krange in self.ranges]

        return KeySetPB(**kwargs)

    def _to_dict(self):
        """Return the state of the keyset as a dict.

        The result can be used to serialize the instance and reconstitute
        it later using :meth:`_from_dict`.

        :rtype: dict
        :returns: state of this instance.
        """
        if self.all_:
            return {"all": True}

        return {
            "keys": self.keys,
            "ranges": [keyrange._to_dict() for keyrange in self.ranges],
        }

    def __eq__(self, other):
        """Compare by serialized state."""
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self._to_dict() == other._to_dict()

    @classmethod
    def _from_dict(cls, mapping):
        """Create an instance from the corresponding state mapping.

        :type mapping: dict
        :param mapping: the instance state.
        """
        if mapping.get("all"):
            return cls(all_=True)

        r_mappings = mapping.get("ranges", ())
        ranges = [KeyRange(**r_mapping) for r_mapping in r_mappings]

        return cls(keys=mapping.get("keys", ()), ranges=ranges)



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Union,
)

if TYPE_CHECKING:
    try:
        from langchain_core import runnables
        from langchain_core import tools as lc_tools
        from langchain_core.language_models import base as lc_language_models

        BaseTool = lc_tools.BaseTool
        BaseLanguageModel = lc_language_models.BaseLanguageModel
        GetSessionHistoryCallable = runnables.history.GetSessionHistoryCallable
        RunnableConfig = runnables.RunnableConfig
        RunnableSerializable = runnables.RunnableSerializable
    except ImportError:
        BaseTool = Any
        BaseLanguageModel = Any
        GetSessionHistoryCallable = Any
        RunnableConfig = Any
        RunnableSerializable = Any

    try:
        from langchain_google_vertexai.functions_utils import _ToolsType

        _ToolLike = _ToolsType
    except ImportError:
        _ToolLike = Any

    try:
        from opentelemetry.sdk import trace

        TracerProvider = trace.TracerProvider
        SpanProcessor = trace.SpanProcessor
        SynchronousMultiSpanProcessor = trace.SynchronousMultiSpanProcessor
    except ImportError:
        TracerProvider = Any
        SpanProcessor = Any
        SynchronousMultiSpanProcessor = Any


def _default_runnable_kwargs(has_history: bool) -> Mapping[str, Any]:
    # https://github.com/langchain-ai/langchain/blob/5784dfed001730530637793bea1795d9d5a7c244/libs/core/langchain_core/runnables/history.py#L237-L241
    runnable_kwargs = {
        # input_messages_key (str): Must be specified if the underlying
        # agent accepts a dict as input.
        "input_messages_key": "input",
        # output_messages_key (str): Must be specified if the underlying
        # agent returns a dict as output.
        "output_messages_key": "output",
    }
    if has_history:
        # history_messages_key (str): Must be specified if the underlying
        # agent accepts a dict as input and a separate key for historical
        # messages.
        runnable_kwargs["history_messages_key"] = "history"
    return runnable_kwargs


def _default_output_parser():
    try:
        from langchain.agents.output_parsers.tools import ToolsAgentOutputParser
    except (ModuleNotFoundError, ImportError):
        # Fallback to an older version if needed.
        from langchain.agents.output_parsers.openai_tools import (
            OpenAIToolsAgentOutputParser as ToolsAgentOutputParser,
        )

    return ToolsAgentOutputParser()


def _default_model_builder(
    model_name: str,
    *,
    project: str,
    location: str,
    model_kwargs: Optional[Mapping[str, Any]] = None,
) -> "BaseLanguageModel":
    import vertexai
    from google.cloud.aiplatform import initializer
    from langchain_google_vertexai import ChatVertexAI

    model_kwargs = model_kwargs or {}
    current_project = initializer.global_config.project
    current_location = initializer.global_config.location
    vertexai.init(project=project, location=location)
    model = ChatVertexAI(model_name=model_name, **model_kwargs)
    vertexai.init(project=current_project, location=current_location)
    return model


def _default_runnable_builder(
    model: "BaseLanguageModel",
    *,
    system_instruction: Optional[str] = None,
    tools: Optional[Sequence["_ToolLike"]] = None,
    prompt: Optional["RunnableSerializable"] = None,
    output_parser: Optional["RunnableSerializable"] = None,
    chat_history: Optional["GetSessionHistoryCallable"] = None,
    model_tool_kwargs: Optional[Mapping[str, Any]] = None,
    agent_executor_kwargs: Optional[Mapping[str, Any]] = None,
    runnable_kwargs: Optional[Mapping[str, Any]] = None,
) -> "RunnableSerializable":
    from langchain_core import tools as lc_tools
    from langchain.agents import AgentExecutor
    from langchain.tools.base import StructuredTool

    # The prompt template and runnable_kwargs needs to be customized depending
    # on whether the user intends for the agent to have history. The way the
    # user would reflect that is by setting chat_history (which defaults to
    # None).
    has_history: bool = chat_history is not None
    prompt = prompt or _default_prompt(
        has_history=has_history,
        system_instruction=system_instruction,
    )
    output_parser = output_parser or _default_output_parser()
    model_tool_kwargs = model_tool_kwargs or {}
    agent_executor_kwargs = agent_executor_kwargs or {}
    runnable_kwargs = runnable_kwargs or _default_runnable_kwargs(has_history)
    if tools:
        model = model.bind_tools(tools=tools, **model_tool_kwargs)
    else:
        tools = []
    agent_executor = AgentExecutor(
        agent=prompt | model | output_parser,
        tools=[
            (
                tool
                if isinstance(tool, lc_tools.BaseTool)
                else StructuredTool.from_function(tool)
            )
            for tool in tools
            if isinstance(tool, (Callable, lc_tools.BaseTool))
        ],
        **agent_executor_kwargs,
    )
    if has_history:
        from langchain_core.runnables.history import RunnableWithMessageHistory

        return RunnableWithMessageHistory(
            runnable=agent_executor,
            get_session_history=chat_history,
            **runnable_kwargs,
        )
    return agent_executor


def _default_prompt(
    has_history: bool,
    system_instruction: Optional[str] = None,
) -> "RunnableSerializable":
    from langchain_core import prompts

    try:
        from langchain.agents.format_scratchpad.tools import format_to_tool_messages
    except (ModuleNotFoundError, ImportError):
        # Fallback to an older version if needed.
        from langchain.agents.format_scratchpad.openai_tools import (
            format_to_openai_tool_messages as format_to_tool_messages,
        )

    system_instructions = []
    if system_instruction:
        system_instructions = [("system", system_instruction)]

    if has_history:
        return {
            "history": lambda x: x["history"],
            "input": lambda x: x["input"],
            "agent_scratchpad": (
                lambda x: format_to_tool_messages(x["intermediate_steps"])
            ),
        } | prompts.ChatPromptTemplate.from_messages(
            system_instructions
            + [
                prompts.MessagesPlaceholder(variable_name="history"),
                ("user", "{input}"),
                prompts.MessagesPlaceholder(variable_name="agent_scratchpad"),
            ]
        )
    else:
        return {
            "input": lambda x: x["input"],
            "agent_scratchpad": (
                lambda x: format_to_tool_messages(x["intermediate_steps"])
            ),
        } | prompts.ChatPromptTemplate.from_messages(
            system_instructions
            + [
                ("user", "{input}"),
                prompts.MessagesPlaceholder(variable_name="agent_scratchpad"),
            ]
        )


def _validate_callable_parameters_are_annotated(callable: Callable):
    """Validates that the parameters of the callable have type annotations.

    This ensures that they can be used for constructing LangChain tools that are
    usable with Gemini function calling.
    """
    import inspect

    parameters = dict(inspect.signature(callable).parameters)
    for name, parameter in parameters.items():
        if parameter.annotation == inspect.Parameter.empty:
            raise TypeError(
                f"Callable={callable.__name__} has untyped input_arg={name}. "
                f"Please specify a type when defining it, e.g. `{name}: str`."
            )


def _validate_tools(tools: Sequence["_ToolLike"]):
    """Validates that the tools are usable for tool calling."""
    for tool in tools:
        if isinstance(tool, Callable):
            _validate_callable_parameters_are_annotated(tool)


def _override_active_span_processor(
    tracer_provider: "TracerProvider",
    active_span_processor: "SynchronousMultiSpanProcessor",
):
    """Overrides the active span processor.

    When working with multiple LangchainAgents in the same environment,
    it's crucial to manage trace exports carefully.
    Each agent needs its own span processor tied to a unique project ID.
    While we add a new span processor for each agent, this can lead to
    unexpected behavior.
    For instance, with two agents linked to different projects, traces from the
    second agent might be sent to both projects.
    To prevent this and guarantee traces go to the correct project, we overwrite
    the active span processor whenever a new LangchainAgent is created.

    Args:
        tracer_provider (TracerProvider):
            The tracer provider to use for the project.
        active_span_processor (SynchronousMultiSpanProcessor):
            The active span processor overrides the tracer provider's
            active span processor.
    """
    if tracer_provider._active_span_processor:
        tracer_provider._active_span_processor.shutdown()
    tracer_provider._active_span_processor = active_span_processor


class LangchainAgent:
    """A Langchain Agent.

    See https://cloud.google.com/vertex-ai/generative-ai/docs/agent-engine/develop/langchain
    for details.
    """

    agent_framework = "langchain"

    def __init__(
        self,
        model: str,
        *,
        system_instruction: Optional[str] = None,
        prompt: Optional["RunnableSerializable"] = None,
        tools: Optional[Sequence["_ToolLike"]] = None,
        output_parser: Optional["RunnableSerializable"] = None,
        chat_history: Optional["GetSessionHistoryCallable"] = None,
        model_kwargs: Optional[Mapping[str, Any]] = None,
        model_tool_kwargs: Optional[Mapping[str, Any]] = None,
        agent_executor_kwargs: Optional[Mapping[str, Any]] = None,
        runnable_kwargs: Optional[Mapping[str, Any]] = None,
        model_builder: Optional[Callable] = None,
        runnable_builder: Optional[Callable] = None,
        enable_tracing: bool = False,
    ):
        """Initializes the LangchainAgent.

        Under-the-hood, assuming .set_up() is called, this will correspond to

        ```
        model = model_builder(model_name=model, model_kwargs=model_kwargs)
        runnable = runnable_builder(
            prompt=prompt,
            model=model,
            tools=tools,
            output_parser=output_parser,
            chat_history=chat_history,
            agent_executor_kwargs=agent_executor_kwargs,
            runnable_kwargs=runnable_kwargs,
        )
        ```

        When everything is based on their default values, this corresponds to
        ```
        # model_builder
        from langchain_google_vertexai import ChatVertexAI
        llm = ChatVertexAI(model_name=model, **model_kwargs)

        # runnable_builder
        from langchain import agents
        from langchain_core.runnables.history import RunnableWithMessageHistory
        llm_with_tools = llm.bind_tools(tools=tools, **model_tool_kwargs)
        agent_executor = agents.AgentExecutor(
            agent=prompt | llm_with_tools | output_parser,
            tools=tools,
            **agent_executor_kwargs,
        )
        runnable = RunnableWithMessageHistory(
            runnable=agent_executor,
            get_session_history=chat_history,
            **runnable_kwargs,
        )
        ```

        Args:
            model (str):
                Optional. The name of the model (e.g. "gemini-1.0-pro").
            system_instruction (str):
                Optional. The system instruction to use for the agent. This
                argument should not be specified if `prompt` is specified.
            prompt (langchain_core.runnables.RunnableSerializable):
                Optional. The prompt template for the model. Defaults to a
                ChatPromptTemplate.
            tools (Sequence[langchain_core.tools.BaseTool, Callable]):
                Optional. The tools for the agent to be able to use. All input
                callables (e.g. function or class method) will be converted
                to a langchain.tools.base.StructuredTool. Defaults to None.
            output_parser (langchain_core.runnables.RunnableSerializable):
                Optional. The output parser for the model. Defaults to an
                output parser that works with Gemini function-calling.
            chat_history (langchain_core.runnables.history.GetSessionHistoryCallable):
                Optional. Callable that returns a new BaseChatMessageHistory.
                Defaults to None, i.e. chat_history is not preserved.
            model_kwargs (Mapping[str, Any]):
                Optional. Additional keyword arguments for the constructor of
                chat_models.ChatVertexAI. An example would be
                ```
                {
                    # temperature (float): Sampling temperature, it controls the
                    # degree of randomness in token selection.
                    "temperature": 0.28,
                    # max_output_tokens (int): Token limit determines the
                    # maximum amount of text output from one prompt.
                    "max_output_tokens": 1000,
                    # top_p (float): Tokens are selected from most probable to
                    # least, until the sum of their probabilities equals the
                    # top_p value.
                    "top_p": 0.95,
                    # top_k (int): How the model selects tokens for output, the
                    # next token is selected from among the top_k most probable
                    # tokens.
                    "top_k": 40,
                }
                ```
            model_tool_kwargs (Mapping[str, Any]):
                Optional. Additional keyword arguments when binding tools to the
                model using `model.bind_tools()`.
            agent_executor_kwargs (Mapping[str, Any]):
                Optional. Additional keyword arguments for the constructor of
                langchain.agents.AgentExecutor. An example would be
                ```
                {
                    # Whether to return the agent's trajectory of intermediate
                    # steps at the end in addition to the final output.
                    "return_intermediate_steps": False,
                    # The maximum number of steps to take before ending the
                    # execution loop.
                    "max_iterations": 15,
                    # The method to use for early stopping if the agent never
                    # returns `AgentFinish`. Either 'force' or 'generate'.
                    "early_stopping_method": "force",
                    # How to handle errors raised by the agent's output parser.
                    # Defaults to `False`, which raises the error.
                    "handle_parsing_errors": False,
                }
                ```
            runnable_kwargs (Mapping[str, Any]):
                Optional. Additional keyword arguments for the constructor of
                langchain.runnables.history.RunnableWithMessageHistory if
                chat_history is specified. If chat_history is None, this will be
                ignored.
            model_builder (Callable):
                Optional. Callable that returns a new language model. Defaults
                to a a callable that returns ChatVertexAI based on `model`,
                `model_kwargs` and the parameters in `vertexai.init`.
            runnable_builder (Callable):
                Optional. Callable that returns a new runnable. This can be used
                for customizing the orchestration logic of the Agent based on
                the model returned by `model_builder` and the rest of the input
                arguments.
            enable_tracing (bool):
                Optional. Whether to enable tracing in Cloud Trace. Defaults to
                False.

        Raises:
            ValueError: If both `prompt` and `system_instruction` are specified.
            TypeError: If there is an invalid tool (e.g. function with an input
            that did not specify its type).
        """
        from google.cloud.aiplatform import initializer

        self._project = initializer.global_config.project
        self._location = initializer.global_config.location
        self._tools = []
        if tools:
            # We validate tools at initialization for actionable feedback before
            # they are deployed.
            _validate_tools(tools)
            self._tools = tools
        if prompt and system_instruction:
            raise ValueError(
                "Only one of `prompt` or `system_instruction` should be specified. "
                "Consider incorporating the system instruction into the prompt "
                "rather than passing it separately as an argument."
            )
        self._model_name = model
        self._system_instruction = system_instruction
        self._prompt = prompt
        self._output_parser = output_parser
        self._chat_history = chat_history
        self._model_kwargs = model_kwargs
        self._model_tool_kwargs = model_tool_kwargs
        self._agent_executor_kwargs = agent_executor_kwargs
        self._runnable_kwargs = runnable_kwargs
        self._model = None
        self._model_builder = model_builder
        self._runnable = None
        self._runnable_builder = runnable_builder
        self._instrumentor = None
        self._enable_tracing = enable_tracing

    def set_up(self):
        """Sets up the agent for execution of queries at runtime.

        It initializes the model, binds the model with tools, and connects it
        with the prompt template and output parser.

        This method should not be called for an object that being passed to
        the ReasoningEngine service for deployment, as it initializes clients
        that can not be serialized.
        """
        if self._enable_tracing:
            from vertexai.reasoning_engines import _utils

            cloud_trace_exporter = _utils._import_cloud_trace_exporter_or_warn()
            cloud_trace_v2 = _utils._import_cloud_trace_v2_or_warn()
            openinference_langchain = _utils._import_openinference_langchain_or_warn()
            opentelemetry = _utils._import_opentelemetry_or_warn()
            opentelemetry_sdk_trace = _utils._import_opentelemetry_sdk_trace_or_warn()
            if all(
                (
                    cloud_trace_exporter,
                    cloud_trace_v2,
                    openinference_langchain,
                    opentelemetry,
                    opentelemetry_sdk_trace,
                )
            ):
                import google.auth

                credentials, _ = google.auth.default()
                span_exporter = cloud_trace_exporter.CloudTraceSpanExporter(
                    project_id=self._project,
                    client=cloud_trace_v2.TraceServiceClient(
                        credentials=credentials.with_quota_project(self._project),
                    ),
                )
                span_processor: SpanProcessor = (
                    opentelemetry_sdk_trace.export.SimpleSpanProcessor(
                        span_exporter=span_exporter,
                    )
                )
                tracer_provider: TracerProvider = (
                    opentelemetry.trace.get_tracer_provider()
                )
                # Get the appropriate tracer provider:
                # 1. If _TRACER_PROVIDER is already set, use that.
                # 2. Otherwise, if the OTEL_PYTHON_TRACER_PROVIDER environment
                # variable is set, use that.
                # 3. As a final fallback, use _PROXY_TRACER_PROVIDER.
                # If none of the above is set, we log a warning, and
                # create a tracer provider.
                if not tracer_provider:
                    from google.cloud.aiplatform import base

                    _LOGGER = base.Logger(__name__)
                    _LOGGER.warning(
                        "No tracer provider. By default, "
                        "we should get one of the following providers: "
                        "OTEL_PYTHON_TRACER_PROVIDER, _TRACER_PROVIDER, "
                        "or _PROXY_TRACER_PROVIDER."
                    )
                    tracer_provider = opentelemetry_sdk_trace.TracerProvider()
                    opentelemetry.trace.set_tracer_provider(tracer_provider)
                # Avoids AttributeError:
                # 'ProxyTracerProvider' and 'NoOpTracerProvider' objects has no
                # attribute 'add_span_processor'.
                if _utils.is_noop_or_proxy_tracer_provider(tracer_provider):
                    tracer_provider = opentelemetry_sdk_trace.TracerProvider()
                    opentelemetry.trace.set_tracer_provider(tracer_provider)
                # Avoids OpenTelemetry client already exists error.
                _override_active_span_processor(
                    tracer_provider,
                    opentelemetry_sdk_trace.SynchronousMultiSpanProcessor(),
                )
                tracer_provider.add_span_processor(span_processor)
                # Keep the instrumentation up-to-date.
                # When creating multiple LangchainAgents,
                # we need to keep the instrumentation up-to-date.
                # We deliberately override the instrument each time,
                # so that if different agents end up using different
                # instrumentations, we guarantee that the user is always
                # working with the most recent agent's instrumentation.
                self._instrumentor = openinference_langchain.LangChainInstrumentor()
                if self._instrumentor.is_instrumented_by_opentelemetry:
                    self._instrumentor.uninstrument()
                self._instrumentor.instrument()
            else:
                from google.cloud.aiplatform import base

                _LOGGER = base.Logger(__name__)
                _LOGGER.warning(
                    "enable_tracing=True but proceeding with tracing disabled "
                    "because not all packages for tracing have been installed"
                )
        model_builder = self._model_builder or _default_model_builder
        self._model = model_builder(
            model_name=self._model_name,
            model_kwargs=self._model_kwargs,
            project=self._project,
            location=self._location,
        )
        runnable_builder = self._runnable_builder or _default_runnable_builder
        self._runnable = runnable_builder(
            prompt=self._prompt,
            model=self._model,
            tools=self._tools,
            system_instruction=self._system_instruction,
            output_parser=self._output_parser,
            chat_history=self._chat_history,
            model_tool_kwargs=self._model_tool_kwargs,
            agent_executor_kwargs=self._agent_executor_kwargs,
            runnable_kwargs=self._runnable_kwargs,
        )

    def clone(self) -> "LangchainAgent":
        """Returns a clone of the LangchainAgent."""
        import copy

        return LangchainAgent(
            model=self._model_name,
            system_instruction=self._system_instruction,
            prompt=copy.deepcopy(self._prompt),
            tools=copy.deepcopy(self._tools),
            output_parser=copy.deepcopy(self._output_parser),
            chat_history=copy.deepcopy(self._chat_history),
            model_kwargs=copy.deepcopy(self._model_kwargs),
            model_tool_kwargs=copy.deepcopy(self._model_tool_kwargs),
            agent_executor_kwargs=copy.deepcopy(self._agent_executor_kwargs),
            runnable_kwargs=copy.deepcopy(self._runnable_kwargs),
            model_builder=self._model_builder,
            runnable_builder=self._runnable_builder,
            enable_tracing=self._enable_tracing,
        )

    def query(
        self,
        *,
        input: Union[str, Mapping[str, Any]],
        config: Optional["RunnableConfig"] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Queries the Agent with the given input and config.

        Args:
            input (Union[str, Mapping[str, Any]]):
                Required. The input to be passed to the Agent.
            config (langchain_core.runnables.RunnableConfig):
                Optional. The config (if any) to be used for invoking the Agent.
            **kwargs:
                Optional. Any additional keyword arguments to be passed to the
                `.invoke()` method of the corresponding AgentExecutor.

        Returns:
            The output of querying the Agent with the given input and config.
        """
        from langchain.load import dump as langchain_load_dump

        if isinstance(input, str):
            input = {"input": input}
        if not self._runnable:
            self.set_up()
        return langchain_load_dump.dumpd(
            self._runnable.invoke(input=input, config=config, **kwargs)
        )

    def stream_query(
        self,
        *,
        input: Union[str, Mapping[str, Any]],
        config: Optional["RunnableConfig"] = None,
        **kwargs,
    ) -> Iterable[Any]:
        """Stream queries the Agent with the given input and config.

        Args:
            input (Union[str, Mapping[str, Any]]):
                Required. The input to be passed to the Agent.
            config (langchain_core.runnables.RunnableConfig):
                Optional. The config (if any) to be used for invoking the Agent.
            **kwargs:
                Optional. Any additional keyword arguments to be passed to the
                `.invoke()` method of the corresponding AgentExecutor.

        Yields:
            The output of querying the Agent with the given input and config.
        """
        from langchain.load import dump as langchain_load_dump

        if isinstance(input, str):
            input = {"input": input}
        if not self._runnable:
            self.set_up()
        for chunk in self._runnable.stream(input=input, config=config, **kwargs):
            yield langchain_load_dump.dumpd(chunk)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains a set of AstDecorator classes, which define the behavior of CrossSync decorators.
Each AstDecorator class is used through @CrossSync.<decorator_name>
"""
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    import ast
    from typing import Callable, Any


class AstDecorator:
    """
    Helper class for CrossSync decorators used for guiding ast transformations.

    AstDecorators are accessed in two ways:
    1. The decorations are used directly as method decorations in the async client,
        wrapping existing classes and methods
    2. The decorations are read back when processing the AST transformations when
        generating sync code.

    This class allows the same decorator to be used in both contexts.

    Typically, AstDecorators act as a no-op in async code, and the arguments simply
    provide configuration guidance for the sync code generation.
    """

    @classmethod
    def decorator(cls, *args, **kwargs) -> Callable[..., Any]:
        """
        Provides a callable that can be used as a decorator function in async code

        AstDecorator.decorate is called by CrossSync when attaching decorators to
        the CrossSync class.

        This method creates a new instance of the class, using the arguments provided
        to the decorator, and defers to the async_decorator method of the instance
        to build the wrapper function.

        Arguments:
            *args: arguments to the decorator
            **kwargs: keyword arguments to the decorator
        """
        # decorators with no arguments will provide the function to be wrapped
        # as the first argument. Pull it out if it exists
        func = None
        if len(args) == 1 and callable(args[0]):
            func = args[0]
            args = args[1:]
        # create new AstDecorator instance from given decorator arguments
        new_instance = cls(*args, **kwargs)
        # build wrapper
        wrapper = new_instance.async_decorator()
        if wrapper is None:
            # if no wrapper, return no-op decorator
            return func or (lambda f: f)
        elif func:
            # if we can, return single wrapped function
            return wrapper(func)
        else:
            # otherwise, return decorator function
            return wrapper

    def async_decorator(self) -> Callable[..., Any] | None:
        """
        Decorator to apply the async_impl decorator to the wrapped function

        Default implementation is a no-op
        """
        return None

    def sync_ast_transform(
        self, wrapped_node: ast.AST, transformers_globals: dict[str, Any]
    ) -> ast.AST | None:
        """
        When this decorator is encountered in the ast during sync generation, this method is called
        to transform the wrapped node.

        If None is returned, the node will be dropped from the output file.

        Args:
            wrapped_node: ast node representing the wrapped function or class that is being wrapped
            transformers_globals: the set of globals() from the transformers module. This is used to access
                ast transformer classes that live outside the main codebase
        Returns:
            transformed ast node, or None if the node should be dropped
        """
        return wrapped_node

    @classmethod
    def get_for_node(cls, node: ast.Call | ast.Attribute | ast.Name) -> "AstDecorator":
        """
        Build an AstDecorator instance from an ast decorator node

        The right subclass is found by comparing the string representation of the
        decorator name to the class name. (Both names are converted to lowercase and
        underscores are removed for comparison). If a matching subclass is found,
        a new instance is created with the provided arguments.

        Args:
            node: ast.Call node representing the decorator
        Returns:
            AstDecorator instance corresponding to the decorator
        Raises:
            ValueError: if the decorator cannot be parsed
        """
        import ast

        # expect decorators in format @CrossSync.<decorator_name>
        # (i.e. should be an ast.Call or an ast.Attribute)
        root_attr = node.func if isinstance(node, ast.Call) else node
        if not isinstance(root_attr, ast.Attribute):
            raise ValueError("Unexpected decorator format")
        # extract the module and decorator names
        if "CrossSync" in ast.dump(root_attr):
            decorator_name = root_attr.attr
            got_kwargs: dict[str, Any] = (
                {str(kw.arg): cls._convert_ast_to_py(kw.value) for kw in node.keywords}
                if hasattr(node, "keywords")
                else {}
            )
            got_args = (
                [cls._convert_ast_to_py(arg) for arg in node.args]
                if hasattr(node, "args")
                else []
            )
            # convert to standardized representation
            formatted_name = decorator_name.replace("_", "").lower()
            for subclass in cls.get_subclasses():
                if subclass.__name__.lower() == formatted_name:
                    return subclass(*got_args, **got_kwargs)
            raise ValueError(f"Unknown decorator encountered: {decorator_name}")
        else:
            raise ValueError("Not a CrossSync decorator")

    @classmethod
    def get_subclasses(cls) -> Iterable[type["AstDecorator"]]:
        """
        Get all subclasses of AstDecorator

        Returns:
            list of all subclasses of AstDecorator
        """
        for subclass in cls.__subclasses__():
            yield from subclass.get_subclasses()
            yield subclass

    @classmethod
    def _convert_ast_to_py(cls, ast_node: ast.expr | None) -> Any:
        """
        Helper to convert ast primitives to python primitives. Used when unwrapping arguments
        """
        import ast

        if ast_node is None:
            return None
        if isinstance(ast_node, ast.Constant):
            return ast_node.value
        if isinstance(ast_node, ast.List):
            return [cls._convert_ast_to_py(node) for node in ast_node.elts]
        if isinstance(ast_node, ast.Tuple):
            return tuple(cls._convert_ast_to_py(node) for node in ast_node.elts)
        if isinstance(ast_node, ast.Dict):
            return {
                cls._convert_ast_to_py(k): cls._convert_ast_to_py(v)
                for k, v in zip(ast_node.keys, ast_node.values)
            }
        # unsupported node type
        return ast_node


class ConvertClass(AstDecorator):
    """
    Class decorator for guiding generation of sync classes

    Args:
        sync_name: use a new name for the sync class
        replace_symbols: a dict of symbols and replacements to use when generating sync class
        docstring_format_vars: a dict of variables to replace in the docstring
        rm_aio: if True, automatically strip all asyncio keywords from method. If false,
            only keywords wrapped in CrossSync.rm_aio() calls to be removed.
        add_mapping_for_name: when given, will add a new attribute to CrossSync,
            so the original class and its sync version can be accessed from CrossSync.<name>
    """

    def __init__(
        self,
        sync_name: str | None = None,
        *,
        replace_symbols: dict[str, str] | None = None,
        docstring_format_vars: dict[str, tuple[str | None, str | None]] | None = None,
        rm_aio: bool = False,
        add_mapping_for_name: str | None = None,
    ):
        self.sync_name = sync_name
        self.replace_symbols = replace_symbols
        docstring_format_vars = docstring_format_vars or {}
        self.async_docstring_format_vars = {
            k: v[0] or "" for k, v in docstring_format_vars.items()
        }
        self.sync_docstring_format_vars = {
            k: v[1] or "" for k, v in docstring_format_vars.items()
        }
        self.rm_aio = rm_aio
        self.add_mapping_for_name = add_mapping_for_name

    def async_decorator(self):
        """
        Use async decorator as a hook to update CrossSync mappings
        """
        from .cross_sync import CrossSync

        if not self.add_mapping_for_name and not self.async_docstring_format_vars:
            # return None if no changes needed
            return None

        new_mapping = self.add_mapping_for_name

        def decorator(cls):
            if new_mapping:
                CrossSync.add_mapping(new_mapping, cls)
            if self.async_docstring_format_vars:
                cls.__doc__ = cls.__doc__.format(**self.async_docstring_format_vars)
            return cls

        return decorator

    def sync_ast_transform(self, wrapped_node, transformers_globals):
        """
        Transform async class into sync copy
        """
        import ast
        import copy

        # copy wrapped node
        wrapped_node = copy.deepcopy(wrapped_node)
        # update name
        if self.sync_name:
            wrapped_node.name = self.sync_name
        # strip CrossSync decorators
        if hasattr(wrapped_node, "decorator_list"):
            wrapped_node.decorator_list = [
                d for d in wrapped_node.decorator_list if "CrossSync" not in ast.dump(d)
            ]
        else:
            wrapped_node.decorator_list = []
        # strip async keywords if specified
        if self.rm_aio:
            wrapped_node = transformers_globals["AsyncToSync"]().visit(wrapped_node)
        # add mapping decorator if needed
        if self.add_mapping_for_name:
            wrapped_node.decorator_list.append(
                ast.Call(
                    func=ast.Attribute(
                        value=ast.Name(id="CrossSync", ctx=ast.Load()),
                        attr="add_mapping_decorator",
                        ctx=ast.Load(),
                    ),
                    args=[
                        ast.Constant(value=self.add_mapping_for_name),
                    ],
                    keywords=[],
                )
            )
        # replace symbols if specified
        if self.replace_symbols:
            wrapped_node = transformers_globals["SymbolReplacer"](
                self.replace_symbols
            ).visit(wrapped_node)
        # update docstring if specified
        if self.sync_docstring_format_vars:
            docstring = ast.get_docstring(wrapped_node)
            if docstring:
                wrapped_node.body[0].value = ast.Constant(
                    value=docstring.format(**self.sync_docstring_format_vars)
                )
        return wrapped_node


class Convert(ConvertClass):
    """
    Method decorator to mark async methods to be converted to sync methods

    Args:
        sync_name: use a new name for the sync method
        replace_symbols: a dict of symbols and replacements to use when generating sync method
        docstring_format_vars: a dict of variables to replace in the docstring
        rm_aio: if True, automatically strip all asyncio keywords from method. If False,
            only the signature `async def` is stripped. Other keywords must be wrapped in
            CrossSync.rm_aio() calls to be removed.
    """

    def __init__(
        self,
        sync_name: str | None = None,
        *,
        replace_symbols: dict[str, str] | None = None,
        docstring_format_vars: dict[str, tuple[str | None, str | None]] | None = None,
        rm_aio: bool = True,
    ):
        super().__init__(
            sync_name=sync_name,
            replace_symbols=replace_symbols,
            docstring_format_vars=docstring_format_vars,
            rm_aio=rm_aio,
            add_mapping_for_name=None,
        )

    def sync_ast_transform(self, wrapped_node, transformers_globals):
        """
        Transform async method into sync
        """
        import ast

        # replace async function with sync function
        converted = ast.copy_location(
            ast.FunctionDef(
                wrapped_node.name,
                wrapped_node.args,
                wrapped_node.body,
                wrapped_node.decorator_list
                if hasattr(wrapped_node, "decorator_list")
                else [],
                wrapped_node.returns if hasattr(wrapped_node, "returns") else None,
            ),
            wrapped_node,
        )
        # transform based on arguments
        return super().sync_ast_transform(converted, transformers_globals)


class Drop(AstDecorator):
    """
    Method decorator to drop methods or classes from the sync output
    """

    def sync_ast_transform(self, wrapped_node, transformers_globals):
        """
        Drop from sync output
        """
        return None


class Pytest(AstDecorator):
    """
    Used in place of pytest.mark.asyncio to mark tests

    When generating sync version, also runs rm_aio to remove async keywords from
    entire test function

    Args:
        rm_aio: if True, automatically strip all asyncio keywords from test code.
            Defaults to True, to simplify test code generation.
    """

    def __init__(self, rm_aio=True):
        self.rm_aio = rm_aio

    def async_decorator(self):
        import pytest

        return pytest.mark.asyncio

    def sync_ast_transform(self, wrapped_node, transformers_globals):
        """
        convert async to sync
        """
        import ast

        # always convert method to sync
        converted = ast.copy_location(
            ast.FunctionDef(
                wrapped_node.name,
                wrapped_node.args,
                wrapped_node.body,
                wrapped_node.decorator_list
                if hasattr(wrapped_node, "decorator_list")
                else [],
                wrapped_node.returns if hasattr(wrapped_node, "returns") else None,
            ),
            wrapped_node,
        )
        # convert entire body to sync if rm_aio is set
        if self.rm_aio:
            converted = transformers_globals["AsyncToSync"]().visit(converted)
        return converted


class PytestFixture(AstDecorator):
    """
    Used in place of pytest.fixture or pytest.mark.asyncio to mark fixtures

    Args:
        *args: all arguments to pass to pytest.fixture
        **kwargs: all keyword arguments to pass to pytest.fixture
    """

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def async_decorator(self):
        import pytest_asyncio  # type: ignore

        return lambda f: pytest_asyncio.fixture(*self._args, **self._kwargs)(f)

    def sync_ast_transform(self, wrapped_node, transformers_globals):
        import ast
        import copy

        arg_nodes = [
            a if isinstance(a, ast.expr) else ast.Constant(value=a) for a in self._args
        ]
        kwarg_nodes = []
        for k, v in self._kwargs.items():
            if not isinstance(v, ast.expr):
                v = ast.Constant(value=v)
            kwarg_nodes.append(ast.keyword(arg=k, value=v))

        new_node = copy.deepcopy(wrapped_node)
        if not hasattr(new_node, "decorator_list"):
            new_node.decorator_list = []
        new_node.decorator_list.append(
            ast.Call(
                func=ast.Attribute(
                    value=ast.Name(id="pytest", ctx=ast.Load()),
                    attr="fixture",
                    ctx=ast.Load(),
                ),
                args=arg_nodes,
                keywords=kwarg_nodes,
            )
        )
        return new_node



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore


__protobuf__ = proto.module(
    package="google.bigtable.v2",
    manifest={
        "FeatureFlags",
    },
)


class FeatureFlags(proto.Message):
    r"""Feature flags supported or enabled by a client. This is intended to
    be sent as part of request metadata to assure the server that
    certain behaviors are safe to enable. This proto is meant to be
    serialized and websafe-base64 encoded under the
    ``bigtable-features`` metadata key. The value will remain constant
    for the lifetime of a client and due to HTTP2's HPACK compression,
    the request overhead will be tiny. This is an internal
    implementation detail and should not be used by end users directly.

    Attributes:
        reverse_scans (bool):
            Notify the server that the client supports
            reverse scans. The server will reject
            ReadRowsRequests with the reverse bit set when
            this is absent.
        mutate_rows_rate_limit (bool):
            Notify the server that the client enables
            batch write flow control by requesting
            RateLimitInfo from MutateRowsResponse. Due to
            technical reasons, this disables partial
            retries.
        mutate_rows_rate_limit2 (bool):
            Notify the server that the client enables
            batch write flow control by requesting
            RateLimitInfo from MutateRowsResponse. With
            partial retries enabled.
        last_scanned_row_responses (bool):
            Notify the server that the client supports the
            last_scanned_row field in ReadRowsResponse for long-running
            scans.
        routing_cookie (bool):
            Notify the server that the client supports
            using encoded routing cookie strings to retry
            requests with.
        retry_info (bool):
            Notify the server that the client supports
            using retry info back off durations to retry
            requests with.
        client_side_metrics_enabled (bool):
            Notify the server that the client has client
            side metrics enabled.
        traffic_director_enabled (bool):
            Notify the server that the client using
            Traffic Director endpoint.
        direct_access_requested (bool):
            Notify the server that the client explicitly
            opted in for Direct Access.
        peer_info (bool):
            If the client can support using
            BigtablePeerInfo.
    """

    reverse_scans: bool = proto.Field(
        proto.BOOL,
        number=1,
    )
    mutate_rows_rate_limit: bool = proto.Field(
        proto.BOOL,
        number=3,
    )
    mutate_rows_rate_limit2: bool = proto.Field(
        proto.BOOL,
        number=5,
    )
    last_scanned_row_responses: bool = proto.Field(
        proto.BOOL,
        number=4,
    )
    routing_cookie: bool = proto.Field(
        proto.BOOL,
        number=6,
    )
    retry_info: bool = proto.Field(
        proto.BOOL,
        number=7,
    )
    client_side_metrics_enabled: bool = proto.Field(
        proto.BOOL,
        number=8,
    )
    traffic_director_enabled: bool = proto.Field(
        proto.BOOL,
        number=9,
    )
    direct_access_requested: bool = proto.Field(
        proto.BOOL,
        number=10,
    )
    peer_info: bool = proto.Field(
        proto.BOOL,
        number=11,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=protected-access
"""Classes to support Tuning."""

from typing import Dict, List, Optional, Union

from google.auth import credentials as auth_credentials

from google.cloud import aiplatform
from google.cloud.aiplatform import base as aiplatform_base
from google.cloud.aiplatform import compat
from google.cloud.aiplatform import initializer as aiplatform_initializer
from google.cloud.aiplatform import jobs
from google.cloud.aiplatform import utils as aiplatform_utils
from google.cloud.aiplatform.utils import _ipython_utils
from google.cloud.aiplatform_v1.services import (
    gen_ai_tuning_service as gen_ai_tuning_service_v1,
)
from google.cloud.aiplatform_v1beta1.services import (
    gen_ai_tuning_service as gen_ai_tuning_service_v1beta1,
)
from google.cloud.aiplatform_v1beta1.types import (
    tuning_job as gca_tuning_job_types,
)
from google.cloud.aiplatform_v1beta1 import types as gca_types

from google.rpc import status_pb2  # type: ignore

_LOGGER = aiplatform_base.Logger(__name__)


class SourceModel:
    r"""A model that is used in managed OSS supervised tuning.

    Usage:
        ```
        model = SourceModel(
            base_model="meta/llama3.1-8b", # OSS model name <publisher>/<model_name>
            custom_base_model="gs://user-bucket/custom-weights",
        )
        sft_tuning_job = sft.train(
            source_model=model,
            train_dataset="gs://my-bucket/train.jsonl",
            validation_dataset="gs://my-bucket/validation.jsonl",
            epochs=4,
            tuned_model_display_name="my-tuned-model",
            output_uri="gs://user-bucket/tuned-model"
        )

        while not sft_tuning_job.has_ended:
            time.sleep(60)
            sft_tuning_job.refresh()

        tuned_model = aiplatform.Model(sft_tuning_job.tuned_model_name)
        ```
    """

    def __init__(
        self,
        base_model: str,
        custom_base_model: str = "",
    ):
        r"""Initializes SourceModel."""
        self.base_model = base_model
        self.custom_base_model = custom_base_model


class TuningJobClientWithOverride(aiplatform_utils.ClientWithOverride):
    _is_temporary = True
    _default_version = compat.V1BETA1
    _version_map = (
        (compat.V1, gen_ai_tuning_service_v1.client.GenAiTuningServiceClient),
        (compat.V1BETA1, gen_ai_tuning_service_v1beta1.client.GenAiTuningServiceClient),
    )


class TuningJob(aiplatform_base._VertexAiResourceNounPlus):
    """Represents a TuningJob that runs with Google owned models."""

    _resource_noun = "tuningJobs"
    _getter_method = "get_tuning_job"
    _list_method = "list_tuning_jobs"
    _cancel_method = "cancel_tuning_job"
    _delete_method = "delete_tuning_job"
    _parse_resource_name_method = "parse_tuning_job_path"
    _format_resource_name_method = "tuning_job_path"
    _job_type = "tuning/tuningJob"
    _has_displayed_experiments_button = False

    client_class = TuningJobClientWithOverride

    _gca_resource: gca_tuning_job_types.TuningJob
    api_client: gen_ai_tuning_service_v1beta1.client.GenAiTuningServiceClient

    def __init__(self, tuning_job_name: str):
        super().__init__(resource_name=tuning_job_name)
        self._gca_resource: gca_tuning_job_types.TuningJob = self._get_gca_resource(
            resource_name=tuning_job_name
        )

    def refresh(self) -> "TuningJob":
        """Refreshed the tuning job from the service."""
        self._gca_resource: gca_tuning_job_types.TuningJob = self._get_gca_resource(
            resource_name=self.resource_name
        )
        if self.experiment and not self._has_displayed_experiments_button:
            self._has_displayed_experiments_button = True
            _ipython_utils.display_experiment_button(self.experiment)
        return self

    @property
    def tuned_model_name(self) -> Optional[str]:
        return self._gca_resource.tuned_model.model

    @property
    def tuned_model_endpoint_name(self) -> Optional[str]:
        return self._gca_resource.tuned_model.endpoint

    @property
    def _experiment_name(self) -> Optional[str]:
        return self._gca_resource.experiment

    @property
    def experiment(self) -> Optional[aiplatform.Experiment]:
        if self._experiment_name:
            return aiplatform.Experiment(experiment_name=self._experiment_name)

    @property
    def state(self) -> gca_types.JobState:
        return self._gca_resource.state

    @property
    def service_account(self) -> Optional[str]:
        self._assert_gca_resource_is_available()
        return self._gca_resource.service_account

    @property
    def has_ended(self):
        return self.state in jobs._JOB_COMPLETE_STATES

    @property
    def has_succeeded(self):
        return self.state == gca_types.JobState.JOB_STATE_SUCCEEDED

    @property
    def error(self) -> Optional[status_pb2.Status]:
        return self._gca_resource.error

    @property
    def tuning_data_statistics(self) -> gca_tuning_job_types.TuningDataStats:
        return self._gca_resource.tuning_data_stats

    @classmethod
    def _create(
        cls,
        *,
        base_model: Union[str, SourceModel],
        tuning_spec: Union[
            gca_tuning_job_types.SupervisedTuningSpec,
            gca_tuning_job_types.DistillationSpec,
        ],
        tuned_model_display_name: Optional[str] = None,
        description: Optional[str] = None,
        labels: Optional[Dict[str, str]] = None,
        project: Optional[str] = None,
        location: Optional[str] = None,
        credentials: Optional[auth_credentials.Credentials] = None,
        output_uri: Optional[str] = None,
    ) -> "TuningJob":
        r"""Submits TuningJob.

        Args:
            base_model: Model for tuning.
                Supported types: str, SourceModel.

                This field is a member of `oneof`_ ``source_model``.
            tuning_spec: Tuning Spec for Fine Tuning.
                Supported types: SupervisedTuningSpec, DistillationSpec.
            tuned_model_display_name: The display name of the
                [TunedModel][google.cloud.aiplatform.v1.Model]. The name can
                be up to 128 characters long and can consist of any UTF-8
                characters.
            description: The description of the `TuningJob`.
            labels: The labels with user-defined metadata to organize
                [TuningJob][google.cloud.aiplatform.v1.TuningJob] and
                generated resources such as
                [Model][google.cloud.aiplatform.v1.Model] and
                [Endpoint][google.cloud.aiplatform.v1.Endpoint].

                Label keys and values can be no longer than 64 characters
                (Unicode codepoints), can only contain lowercase letters,
                numeric characters, underscores and dashes. International
                characters are allowed.

                See https://goo.gl/xmQnxf for more information and examples
                of labels.
            project: Project to run the tuning job in.
                Overrides project set in aiplatform.init.
            location: Location to run the tuning job in.
                Overrides location set in aiplatform.init.
            credentials: Custom credentials to use to call tuning job service.
                Overrides credentials set in aiplatform.init.
            output_uri: The Google Cloud Storage location to write the artifacts. This is only used for OSS models.

        Returns:
            Submitted TuningJob.

        Raises:
            RuntimeError is tuning_spec kind is unsupported
        """
        _LOGGER.log_create_with_lro(cls)

        if not tuned_model_display_name:
            tuned_model_display_name = cls._generate_display_name()

        gca_tuning_job = gca_tuning_job_types.TuningJob(
            tuned_model_display_name=tuned_model_display_name,
            description=description,
            labels=labels,
            # The tuning_spec one_of is set later.
            output_uri=output_uri,
        )

        if isinstance(tuning_spec, gca_tuning_job_types.SupervisedTuningSpec):
            gca_tuning_job.supervised_tuning_spec = tuning_spec
            if isinstance(base_model, SourceModel):
                gca_tuning_job.base_model = base_model.base_model
                gca_tuning_job.custom_base_model = base_model.custom_base_model
            else:
                gca_tuning_job.base_model = base_model
        elif isinstance(tuning_spec, gca_tuning_job_types.DistillationSpec):
            gca_tuning_job.distillation_spec = tuning_spec
            if isinstance(base_model, SourceModel):
                raise RuntimeError("Distillation is not supported for custom models.")
            gca_tuning_job.base_model = base_model
        else:
            raise RuntimeError(f"Unsupported tuning_spec kind: {tuning_spec}")

        if aiplatform_initializer.global_config.encryption_spec_key_name:
            gca_tuning_job.encryption_spec.kms_key_name = (
                aiplatform_initializer.global_config.encryption_spec_key_name
            )
        gca_tuning_job.service_account = (
            aiplatform_initializer.global_config.service_account
        )

        tuning_job: TuningJob = cls._construct_sdk_resource_from_gapic(
            gapic_resource=gca_tuning_job,
            project=project,
            location=location,
            credentials=credentials,
        )

        parent = aiplatform_initializer.global_config.common_location_path(
            project=project, location=location
        )

        created_gca_tuning_job = tuning_job.api_client.create_tuning_job(
            parent=parent,
            tuning_job=gca_tuning_job,
        )
        tuning_job._gca_resource = created_gca_tuning_job

        _LOGGER.log_create_complete(
            cls=cls,
            resource=created_gca_tuning_job,
            variable_name="tuning_job",
            module_name="sft",
        )
        _LOGGER.info(f"View Tuning Job:\n{tuning_job._dashboard_url()}")
        if tuning_job._experiment_name:
            _LOGGER.info(f"View experiment:\n{tuning_job._experiment.dashboard_url}")

        return tuning_job

    def cancel(self):
        self.api_client.cancel_tuning_job(name=self.resource_name)

    @classmethod
    def list(cls, filter: Optional[str] = None) -> List["TuningJob"]:
        """Lists TuningJobs.

        Args:
            filter: The standard list filter.

        Returns:
            A list of TuningJob objects.
        """
        return cls._list(filter=filter)

    def _dashboard_url(self) -> str:
        """Returns the Google Cloud console URL where job can be viewed."""
        fields = self._parse_resource_name(self.resource_name)
        location = fields.pop("location")
        project = fields.pop("project")
        job = list(fields.values())[0]
        url = f"https://console.cloud.google.com/vertex-ai/generative/language/locations/{location}/tuning/tuningJob/{job}?project={project}"
        return url


def rebase_tuned_model(
    tuned_model_ref: str,
    *,
    # TODO(b/372291558): Add support for overriding tuning job config
    # tuning_job_config: Optional["TuningJob"] = None,
    artifact_destination: Optional[str] = None,
    deploy_to_same_endpoint: Optional[bool] = False,
):
    """Re-runs fine tuning on top of a new foundational model.

    Takes a legacy Tuned GenAI model Reference and creates a TuningJob based
    on a new model.

    Args:
        tuned_model_ref: Required. TunedModel reference to retrieve
            the legacy model information.
        tuning_job_config: The TuningJob to be updated. Users
            can use this TuningJob field to overwrite tuning
            configs.
        artifact_destination: The Google Cloud Storage location to write the artifacts.
        deploy_to_same_endpoint:
            Optional. By default, bison to gemini
            migration will always create new model/endpoint,
            but for gemini-1.0 to gemini-1.5 migration, we
            default deploy to the same endpoint. See details
            in this Section.

    Returns:
        The new TuningJob.
    """
    parent = aiplatform_initializer.global_config.common_location_path(
        project=aiplatform_initializer.global_config.project,
        location=aiplatform_initializer.global_config.location,
    )

    if "/tuningJobs/" in tuned_model_ref:
        gapic_tuned_model_ref = gca_types.TunedModelRef(
            tuning_job=tuned_model_ref,
        )
    elif "/pipelineJobs/" in tuned_model_ref:
        gapic_tuned_model_ref = gca_types.TunedModelRef(
            pipeline_job=tuned_model_ref,
        )
    elif "/models/" in tuned_model_ref:
        gapic_tuned_model_ref = gca_types.TunedModelRef(
            tuned_model=tuned_model_ref,
        )
    else:
        raise ValueError(f"Unsupported tuned_model_ref: {tuned_model_ref}.")

    # gapic_tuning_job_config = tuning_job._gca_resource if tuning_job else None
    gapic_tuning_job_config = None

    gapic_artifact_destination = (
        gca_types.GcsDestination(output_uri_prefix=artifact_destination)
        if artifact_destination
        else None
    )

    api_client: gen_ai_tuning_service_v1beta1.GenAiTuningServiceClient = (
        TuningJob._instantiate_client(
            location=aiplatform_initializer.global_config.location,
            credentials=aiplatform_initializer.global_config.credentials,
        )
    )
    rebase_operation = api_client.rebase_tuned_model(
        gca_types.RebaseTunedModelRequest(
            parent=parent,
            tuned_model_ref=gapic_tuned_model_ref,
            tuning_job=gapic_tuning_job_config,
            artifact_destination=gapic_artifact_destination,
            deploy_to_same_endpoint=deploy_to_same_endpoint,
        )
    )
    _LOGGER.log_create_with_lro(TuningJob, lro=rebase_operation)
    gapic_rebase_tuning_job: gca_types.TuningJob = rebase_operation.result()
    rebase_tuning_job = TuningJob._construct_sdk_resource_from_gapic(
        gapic_resource=gapic_rebase_tuning_job,
    )
    return rebase_tuning_job



# -*- coding: utf-8 -*-

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from google.cloud.logging_v2 import __version__
from google.cloud.logging_v2 import ASCENDING
from google.cloud.logging_v2 import DESCENDING

from google.cloud.logging_v2.client import Client
from google.cloud.logging_v2.entries import logger_name_from_path
from google.cloud.logging_v2.entries import LogEntry
from google.cloud.logging_v2.entries import TextEntry
from google.cloud.logging_v2.entries import StructEntry
from google.cloud.logging_v2.entries import ProtobufEntry
from google.cloud.logging_v2 import handlers
from google.cloud.logging_v2.logger import Logger
from google.cloud.logging_v2.logger import Batch
from google.cloud.logging_v2.metric import Metric
from google.cloud.logging_v2.resource import Resource
from google.cloud.logging_v2.sink import Sink
from google.cloud.logging_v2 import types

__all__ = (
    "__version__",
    "ASCENDING",
    "Batch",
    "Client",
    "DESCENDING",
    "handlers",
    "logger_name_from_path",
    "Logger",
    "LogEntry",
    "Metric",
    "ProtobufEntry",
    "Resource",
    "Sink",
    "StructEntry",
    "TextEntry",
    "types",
)
